/** @file emu.h

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.02.21

	@brief [ sdl emulation i/f ]

	@note
	This code is based on the Common Source Code Project.
	Original Author : Takeda.Toshiya
*/

#ifndef _EMU_H_
#define _EMU_H_

#include "common.h"
#include "logging.h"
#include "utility.h"
#include "config.h"
#include "vm/vm.h"
#include "clocale.h"

// OpenGL
#ifdef USE_OPENGL
#include <SDL_opengl.h>
#endif

#ifndef SCREEN_WIDTH_ASPECT
#define SCREEN_WIDTH_ASPECT SCREEN_WIDTH
#endif
#ifndef SCREEN_HEIGHT_ASPECT
#define SCREEN_HEIGHT_ASPECT SCREEN_HEIGHT
#endif
#ifndef WINDOW_WIDTH
#define WINDOW_WIDTH SCREEN_WIDTH_ASPECT
#endif
#ifndef WINDOW_HEIGHT
#define WINDOW_HEIGHT SCREEN_HEIGHT_ASPECT
#endif

#define SDL_USEREVENT_COMMAND (SDL_USEREVENT + 1)

#define SCREEN_MODE_MAX 24
typedef struct screen_mode_st {
	int disp_no;
	int width;
	int height;
} screen_mode_t;
#define WINDOW_MODE_MAX 8
typedef struct window_mode_st {
	int power;
	int width;
	int height;
} window_mode_t;

#ifdef USE_MEDIA
#define MEDIA_MAX 64
#endif

#ifdef USE_SOCKET
#include <SDL_net.h>
#define SOCKET_MAX 6
#define SOCKET_BUFFER_MAX 0x100000
#define SDL_USEREVENT_SOCKET (SDL_USEREVENT_COMMAND + 1)
#define SDL_USEREVENT_SOCKET_CONNECTED 1
#define SDL_USEREVENT_SOCKET_DISCONNECTED 2
#define SDL_USEREVENT_SOCKET_WRITEABLE 3
#define SDL_USEREVENT_SOCKET_READABLE 4
#define SDL_USEREVENT_SOCKET_ACCEPT 5
#endif

#define KEY_STATUS_SIZE 512
#define KEY_MOD_ALT_KEY 0x0001

#ifdef USE_FD1
// for d88 bank switch
#define MAX_D88_BANKS 50

typedef struct {
	_TCHAR name[20];
	int offset;
} d88_bank_t;

typedef struct {
	_TCHAR path[_MAX_PATH];
	d88_bank_t bank[MAX_D88_BANKS];
	int bank_nums;
	int cur_bank;
	int prev_bank;
} d88_file_t;
#endif

class FIFO;
class FILEIO;
class GUI;
//class LedBox;
class MsgBoard;
class REC_VIDEO;

/**
	@brief emulation i/f
*/
class EMU : public Logging
{
protected:
	VM* vm;
	GUI *gui;
	bool initialized;
private:
	// ----------------------------------------
	// input
	// ----------------------------------------
	/// @name input private methods
	//@{
	void EMU_INPUT();
	void initialize_input();
	void release_input();
	void update_input();

	void initialize_joystick();
	void release_joystick();

	void initialize_mouse(bool enable);
#ifdef USE_MOUSE_ABSOLUTE
	void set_mouse_position();
#endif

	int  translate_global_key(int code);
	//@}
	/// @name input private members
	//@{
	uint8 key_status[KEY_STATUS_SIZE];	// windows key code mapping
	uint8 *vm_key_status;	// vm key scan code mapping (provided from vm)
	int   vm_key_status_size;
#ifdef USE_SHIFT_NUMPAD_KEY
	uint8 key_converted[KEY_STATUS_SIZE];
	bool key_shift_pressed, key_shift_released;
#endif
	bool lost_focus;
	int  key_mod;
	bool pressed_global_key;

	uint32 joy_status[2];	// joystick #1, #2 (b0 = up, b1 = down, b2 = left, b3 = right, b4-b31 = trigger #1-#28
	int joy_num;
	int joy_xmin[2], joy_xmax[2];
	int joy_ymin[2], joy_ymax[2];
	bool joy_enabled[2];
	SDL_Joystick *joy[2];
	bool use_joystick;

	int mouse_status[3];	// x, y, button (b0 = left, b1 = right)
	int mouse_disabled;
#ifdef USE_MOUSE_ABSOLUTE
	int mouse_posx, mouse_posy;
#endif

#ifdef USE_AUTO_KEY
	FIFO* autokey_buffer;
	int autokey_phase, autokey_shift;
	uint16 autokey_code;
	bool autokey_enabled;
#endif
	//@}

private:
	// ----------------------------------------
	// screen
	// ----------------------------------------
	/// @name screen private methods
	//@{
	void EMU_SCREEN();
	void initialize_screen();
	void release_screen();
	void release_screen_on_emu_thread();
	void set_vm_display_size();
	int  adjust_y_position(int h, int y);

#ifdef USE_OPENGL
	void initialize_opengl();
	void release_opengl();

	void create_texture();
	void release_texture();

	void set_opengl_attr();
	void set_opengl_poly(int width, int height);
#endif
	//@}
	/// @name screen private static methods
	//@{
	static int qsort_by_w_h_n(const void *a, const void *b);
	static int qsort_by_n_w_h(const void *a, const void *b);
	//@}
	/// @name screen private members
	//@{
	// screen settings
	SDL_PixelFormat pixel_format;
	SDL_Surface *screen;
	SDL_Surface *srcsurface;
	SDL_Surface *mixsurface;
	Uint32 screen_flags;
	SDL_mutex *mux_update_screen;
	bool screen_changing;
#ifdef USE_SDL2
	SDL_Window *window;
#ifdef USE_OPENGL
	SDL_GLContext glcontext;
#endif
#endif

	/// vm draw area
	int screen_width, screen_height;
	int screen_width_aspect, screen_height_aspect;
	// vm draw area (not used)
//	int window_width, window_height;
	/// minimum window client size
	int min_window_width, min_window_height;
	/// display size (window client size)
	int display_width, display_height;
	/// margin in window client area (window mode only)
	VmRect display_margin;
	/// desktop size on the current monitor
	int desktop_width, desktop_height;
	bool screen_size_changed;

	SDL_Rect rec_video_size[2];
	SDL_Rect rec_video_stretched_size;

	/// vm draw area
	/// if use screen rotation, swap width and height.
	int source_width, source_height;
	int source_width_aspect, source_height_aspect;
	int stretched_width, stretched_height;
	int stretch_pow_x, stretch_pow_y;
//	int screen_dest_x, screen_dest_y;
//	int display_stretched_width, display_stretched_height;
//	int screen_stretched_dest_x, screen_stretched_dest_y;
	int stretched_dest_x, stretched_dest_y;
	int stretched_dest_real_x, stretched_dest_real_y;
	int mixed_width, mixed_height;
	int mixed_max_width, mixed_max_height;	// for OpenGL
	int mixed_dest_x, mixed_dest_y;
	int mixed_width_ratio, mixed_height_ratio;

	// update flags
	bool first_draw_screen;
	bool first_invalidate;
	bool self_invalidate;
	bool skip_frame;
	bool skip_frame_pre;

	// screen buffer
	scrntype* lpBmp;

	SDL_Rect reMix;
	SDL_Rect reSuf;

	// screen mode
	int screen_mode_count;
	screen_mode_t screen_mode[SCREEN_MODE_MAX];

	int window_mode_count;
	window_mode_t window_mode[WINDOW_MODE_MAX];
	int window_mode_power;

	bool now_fullscreen;
	int  prev_window_mode;

	// record video
	REC_VIDEO *rec_video;
	bool *now_recording;

#ifdef USE_OPENGL
	GLuint mix_texture_name;
	SDL_Rect reTex;
	GLdouble src_tex_l, src_tex_t, src_tex_r, src_tex_b; 
	int use_opengl;
#endif
	//@}

	// ----------------------------------------
	// sound
	// ----------------------------------------
	/// @name sound private methods
	//@{
	void EMU_SOUND();
	void initialize_sound(int rate, int samples, int latency);
	void initialize_sound(int rate, int samples);
	void initialize_sound();
	void release_sound();
	void release_sound_on_emu_thread();
	void start_sound();
	void end_sound();
//	void update_sound(int* extra_frames);
	//@}
	/// @name sound private static methods
	//@{
	// callback
	static void update_sound(void *userdata, Uint8 *stream, int len);
	//@}
	/// @name sound private members
	//@{
	SDL_AudioSpec audio_desired, audio_obtained;

	int sound_rate, sound_samples;
	int sound_latency;
	bool sound_ok, now_mute;

	// direct sound
	uint32 sound_prev_time;

	// record sound
	typedef struct {
		uint32 dwRIFF;
		uint32 dwFileSize;
		uint32 dwWAVE;
		uint32 dwfmt_;
		uint32 dwFormatSize;
		uint16 wFormatTag;
		uint16 wChannels;
		uint32 dwSamplesPerSec;
		uint32 dwAvgBytesPerSec;
		uint16 wBlockAlign;
		uint16 wBitsPerSample;
		uint32 dwdata;
		uint32 dwDataLength;
	} wavheader_t;
	FILEIO* rec;
	int rec_bytes;
	bool now_rec_snd;
	//@}

	// ----------------------------------------
	// floppy disk
	// ----------------------------------------
#ifdef USE_FD1
	/// @name floppy disk private methods
	//@{
	void initialize_disk_insert();
	void update_disk_insert();
	//@}
	/// @name floppy disk private members
	//@{
	typedef struct {
		_TCHAR path[_MAX_PATH];
		int offset;
		int wait_count;
		uint32 flags;
	} disk_insert_t;
	disk_insert_t disk_insert[MAX_DRIVE];

	// d88 bank switch
	d88_file_t d88_file[MAX_DRIVE];
	//@}
#endif

	// ----------------------------------------
	// media
	// ----------------------------------------
#ifdef USE_MEDIA
	/// @name media private methods
	//@{
	void initialize_media();
	void release_media();
	//@}
	/// @name media private members
	//@{
	_TCHAR media_path[MEDIA_MAX][_MAX_PATH];
	int media_cnt;
	bool media_playing;
	//@}
#endif

	// ----------------------------------------
	// timer
	// ----------------------------------------
	/// @name timer private methods
	//@{
	void update_timer();
	//@}
	/// @name timer private members
	//@{
//	SYSTEMTIME sTime;
	struct tm sTime;
	//@}

	// ----------------------------------------
	// socket
	// ----------------------------------------
#ifdef USE_SOCKET
	/// @name socket private methods
	//@{
	void EMU_SOCKET();
	void initialize_socket();
	void release_socket();
	void update_socket();
	//@}
	/// @name socket private members
	//@{
//	SOCKET soc[SOCKET_MAX];
	bool is_tcp[SOCKET_MAX];
//	struct sockaddr_in udpaddr[SOCKET_MAX];
	int socket_delay[SOCKET_MAX];
	char *recv_buffer[SOCKET_MAX];
	int recv_r_ptr[SOCKET_MAX], recv_w_ptr[SOCKET_MAX];
	int socket_nums;

	SDLNet_SocketSet socket_set;
	TCPsocket tcp_socket[SOCKET_MAX];
	UDPsocket udp_socket[SOCKET_MAX];
	UDPpacket udp_packet;
	bool is_server[SOCKET_MAX];
	//@}
#endif

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc private methods
	//@{
#ifdef USE_EMU_INHERENT_SPEC
	uint64 update_led();
#endif
	//@}
	/// @name misc private members
	//@{
	_TCHAR app_path[_MAX_PATH];
	_TCHAR ini_path[_MAX_PATH];
	_TCHAR res_path[_MAX_PATH];

#define VM_SYSPAUSE_MASK 0x01
#define VM_USRPAUSE_MASK 0x10

	int vm_pause;

//	LedBox *ledbox;
	MsgBoard *msgboard;
	//@}
//
//
//
public:
	// ----------------------------------------
	// initialize
	// ----------------------------------------
	/// @name initialize
	//@{
	EMU(const _TCHAR *new_app_path, const _TCHAR *new_ini_path, const _TCHAR *new_res_path);
	~EMU();
	void initialize();
	void release();
	void release_on_emu_thread();
	//@}

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc
	//@{
	void    application_path(_TCHAR *path);
	const _TCHAR *application_path();
	void    initialize_path(_TCHAR *path);
	const _TCHAR *initialize_path();
	void    resource_path(_TCHAR *path);
	const _TCHAR *resource_path();

	bool get_pause(int idx);
	void set_pause(int idx, bool val);
	int *get_pause_ptr(void);
	void set_gui(GUI *new_gui);
	GUI *get_gui();
	//@}

	// ----------------------------------------
	// for windows
	// ----------------------------------------
	/// @name drive virtual machine
	//@{
	int frame_interval();
	int run(int split_num);
	void update();
	bool now_skip();
	//@}
	// user interface
	/// @name control menu for ui
	//@{
	void update_config();
	void reset();
#ifdef USE_SPECIAL_RESET
	void special_reset();
	void warm_reset(int onoff);
#endif
	void change_dipswitch(int num);
#ifdef USE_POWER_OFF
	void notify_power_off();
#endif
	void change_cpu_power(int num);
#ifdef USE_EMU_INHERENT_SPEC
	void change_sync_irq();
#endif
#ifdef USE_CART
	void open_cart(_TCHAR* file_path);
	void close_cart();
#endif
#ifdef USE_AUTO_KEY
	bool start_auto_key(const char *str);
	bool open_auto_key(const _TCHAR *path);
	void stop_auto_key();
	bool now_auto_key() {
		return autokey_enabled;
	}
#endif
#ifdef USE_EMU_INHERENT_SPEC
	bool save_state(const _TCHAR* file_path);
	bool load_state(const _TCHAR* file_path);
#ifdef USE_KEY_RECORD
	bool play_reckey(const _TCHAR* file_path);
	bool record_reckey(const _TCHAR* file_path);
	void stop_reckey(bool stop_play = true, bool stop_record = true);
#endif
#endif
	//@}
#ifdef USE_FD1
	/// @name floppy disk menu for ui
	//@{
	bool open_disk_main(int drv, const _TCHAR* file_path, int offset, uint32 flags);
	bool open_disk(int drv, const _TCHAR* file_path, int bank_num, int offset, uint32 flags);
	bool open_disk_by_bank_num(int drv, const _TCHAR* file_path, int bank_num, uint32 flags, bool multiopen);
	bool open_disk_with_sel_bank(int drv, int bank_num);
	void close_disk(int drv);
	int  change_disk(int drv);
	int  get_disk_side(int drv);
	void toggle_disk_write_protect(int drv);
	bool disk_write_protected(int drv);
	bool disk_inserted(int drv);
	bool changed_cur_bank(int drv);
	d88_file_t *get_d88_file(int drv) { return &d88_file[drv]; }
	//@}
#endif
#ifdef USE_QUICKDISK
	void open_quickdisk(_TCHAR* file_path);
	void close_quickdisk();
#endif
#ifdef USE_DATAREC
	/// @name tape menu for ui
	//@{
	bool play_datarec(const _TCHAR* file_path);
	bool rec_datarec(const _TCHAR* file_path);
	void close_datarec();
	void rewind_datarec();
	void fast_forward_datarec();
	void realmode_datarec();
	bool datarec_opened(bool play_mode);
	//@}
#endif
#ifdef USE_DATAREC_BUTTON
	void push_play();
	void push_stop();
#endif
#ifdef USE_MEDIA
	void open_media(_TCHAR* file_path);
	void close_media();
#endif
#ifdef USE_BINARY_FILE1
	void load_binary(int drv, _TCHAR* file_path);
	void save_binary(int drv, _TCHAR* file_path);
#endif
	/// @name screen menu for ui
	//@{
	void change_screen_mode(int mode);
	void change_stretch_screen(int num);
//	void change_cutout_screen();
	int get_window_mode_count() {
		return window_mode_count;
	}
	int get_screen_mode_count() {
		return screen_mode_count;
	}
	bool is_fullscreen() {
		return now_fullscreen;
	}
	window_mode_t *get_window_mode(int num);
	screen_mode_t *get_screen_mode(int num);
	SDL_Rect *get_screen_record_size(int num);
	void change_pixel_aspect(int mode);
	int  get_pixel_aspect_count();
	int  get_pixel_aspect(int mode, int *wratio, int *hratio);
	void capture_screen();
	bool start_rec_video(int type, int fps, bool show_dialog);
	void stop_rec_video();
	void restart_rec_video();
	void record_rec_video();
	void resize_rec_video(int num);
	void change_rec_video_size(int num);
	bool now_rec_video();
	bool rec_video_enabled(int type);
	const _TCHAR **get_rec_video_codec_list(int type);
	const _TCHAR **get_rec_video_quality_list(int type);
#ifdef USE_SCANLINE
	void change_screen_scanline(int num);
#endif
#ifdef USE_EMU_INHERENT_SPEC
#ifdef USE_OPENGL
	void change_screen_use_opengl(int num);
	void change_opengl_attr();
	int now_use_opengl() {
		return use_opengl;
	}
#endif
	void change_screen_afterimage(int num);
	void change_screen_keepimage(int num);
#endif
	//@}
	/// @name sound menu for ui
	//@{
	void start_rec_sound();
	void stop_rec_sound();
	void restart_rec_sound();
	bool now_rec_sound() {
		return now_rec_snd;
	}
	bool *get_now_rec_sound_ptr() {
		return &now_rec_snd;
	}
	void set_volume(int volume);
	//@}
#ifdef USE_PRINTER
	/// @name printer menu for ui
	//@{
	bool save_printer(int dev, const _TCHAR *filename);
	void clear_printer(int dev);
	int  get_printer_buffer_size(int dev);
	uint8* get_printer_buffer(int dev);
//	void send_printer_data(uint8 *datas, int len);
	void enable_printer_direct(int dev);
	bool print_printer(int dev);
	//@}
#endif
#ifdef USE_EMU_INHERENT_SPEC
	/// @name comm menu for ui
	//@{
	void enable_comm_server(int dev);
	void enable_comm_connect(int dev);
	//@}
#endif
#ifdef USE_EMU_INHERENT_SPEC
	/// @name options menu for ui
	//@{
//	void show_led();
//	void inside_led();
//	void move_led();
//	void change_pos_led(int num);
//	int  is_shown_ledbox(int id);
	void show_message_board();
	int  is_shown_message_board();
	void change_use_joypad(int num);
	bool is_enable_joypad(int num);

	void save_keybind();

	void change_archtecture(int id, int num, bool reset);
	//@}
#endif

	//
	/// @name input device procedures for host machine
	//@{
//	void key_down(int code, bool repeat);
	int  key_down(int sym, uint16 mod);
//	void key_up(int code);
	void key_up(int sym, uint16 mod);
	void key_lost_focus() {
		lost_focus = true;
	}
	int  vm_key_down(int code);
	void vm_key_up(int code);
	int  system_key_down(int code);
	void system_key_up(int code);
	bool execute_global_keys(int code, uint32 status);
	bool release_global_keys(int code, uint32 status);
	void post_command_message(int id);
	void translate_sdl_keysym(SDL_Event *e, int *new_type, int *new_scancode, int *new_sym);
	void translate_sdl_keysym(Uint32 e_type, Uint16 e_scancode, SDL_Keycode e_sym, Uint16 e_mod, int *new_type, int *new_scancode, int *new_sym);
	bool translate_key_down(int sym, uint16 mod);

#ifdef USE_BUTTON
	void press_button(int num);
#endif
	void reset_joystick();
	void update_joystick();

	void update_mouse();
	void enable_mouse(int mode);
	void disable_mouse(int mode);
	void toggle_mouse();
	bool get_mouse_enabled();

#ifdef USE_AUTO_KEY
	void parse_auto_key(const char *buf, int size, int &prev_code, int &prev_kana);
#endif
	void update_autokey();
//	void enable_joystick();
	//@}
	/// @name screen device procedures for host machine
	//@{
//	int get_window_width(int mode);
//	int get_window_height(int mode);
	bool create_screen(int width, int height, Uint32 flags);
	void set_display_size(int width, int height, int power, bool window_mode);
//	void change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh);
	void draw_screen();
	bool mix_screen();
	void update_screen();
	void skip_screen(bool val) {
		skip_frame = val;
	}
	void need_update_screen();

	void enum_window_mode(int max_width, int max_height);
	void enum_screen_mode(Uint32 flags);
	void find_screen_mode();
	void set_window(int mode, int cur_width, int cur_height);
	void set_desktop_size(int width, int height) {
		desktop_width = width;
		desktop_height = height;
	}
	void set_display_margin(int left, int top, int right, int bottom) {
		display_margin.left = left;
		display_margin.top = top;
		display_margin.right = right;
		display_margin.bottom = bottom;
	}
	void lock_screen();
	void unlock_screen();
#ifdef USE_BITMAP
	void reload_bitmap() {
		first_invalidate = true;
	}
#endif
	bool create_offlinesurface();

#ifdef USE_OPENGL
	int  get_use_opengl() {
		return use_opengl;
	}
	void set_use_opengl(int val) {
		use_opengl = val;
	}
#endif
	SDL_Surface *get_screen() {
		return screen;
	}
#ifdef USE_SDL2
	SDL_Window *get_window() {
		return window;
	}
#endif
	//@}
	/// @name sound device procedures for host machine
	//@{
	void mute_sound(bool mute);
	//@}
#ifdef USE_SOCKET
	/// @name socket device procedures for host machine
	//@{
//	int get_socket(int ch);
	void *get_socket(int ch);
	void socket_connected(int ch);
	void socket_disconnected(int ch);
	void socket_writeable(int ch);
	void socket_readable(int ch);
	void socket_accept(int ch);
	void send_data(int ch);
	void recv_data(int ch);
	//@}
#endif

	// ----------------------------------------
	// for virtual machine
	// ----------------------------------------

	// power off
	void power_off() {
//		PostMessage(main_window_handle, WM_CLOSE, 0, 0L);
	}

	/// @name input device for vm
	//@{
	uint8* key_buffer() {
		return key_status;
	}
	uint32* joy_buffer() {
		return joy_status;
	}
	int* mouse_buffer() {
		return mouse_status;
	}
	int* get_key_mod_ptr() {
		return &key_mod;
	}
	void recognized_key(uint16 key_code);
	void set_vm_key_buffer(uint8 *buffer, int size) {
		vm_key_status = buffer;
		vm_key_status_size = size;
	}
	//@}
	/// @name screen for vm
	//@{
	scrntype* screen_buffer(int y);
	int screen_buffer_offset();
	void get_rgbformat(uint32 *r_mask, uint32 *g_mask, uint32 *b_mask, uint8 *r_shift, uint8 *g_shift, uint8 *b_shift);
	void get_rgbaformat(uint32 *r_mask, uint32 *g_mask, uint32 *b_mask, uint32 *a_mask, uint8 *r_shift, uint8 *g_shift, uint8 *b_shift, uint8 *a_shift);
	void get_rgbcolor(scrntype pixel, uint8 *r, uint8 *g, uint8 *b);
	scrntype map_rgbcolor(uint8 r, uint8 g, uint8 b);
	bool now_skip_frame();
	//@}
	/// @name sound for vm
	//@{
	void record_rec_sound(uint8 *buffer, int samples);
	void record_rec_sound(int16 *buffer, int samples);
	void record_rec_sound(int32 *buffer, int samples);
	void lock_sound_buffer();
	void unlock_sound_buffer();
	//@}
	/// @name timer for vm
	//@{
	void get_timer(int time[]);
	//@}
#ifdef USE_MEDIA
	/// @name media for vm
	//@{
	int media_count();
	void play_media(int trk);
	void stop_media();
	//@}
#endif
#ifdef USE_SOCKET
	/// @name socket for vm
	//@{
	bool init_socket_tcp(int ch, bool server = false);
	bool init_socket_udp(int ch, bool server = false);
	bool connect_socket(int ch, uint32 ipaddr, int port, bool server = false);
	bool connect_socket(int ch, const _TCHAR *hostname, int port, bool server = false);
	bool connect_socket(int ch, IPaddress *ip, bool server = false);
	bool get_ipaddr(const _TCHAR *hostname, int port, IPaddress *ip);
	void disconnect_socket(int ch);
	int  get_socket_channel();
	bool listen_socket(int ch);
	void send_data_tcp(int ch);
	void send_data_udp(int ch, uint32 ipaddr, int port);
	void send_data_udp(int ch, const _TCHAR *hostname, int port);
	//@}
#endif
#ifdef USE_EMU_INHERENT_SPEC
	/// @name send message to ui from vm
	//@{
	void update_ui(int flags);
	//@}
//	void send_comm_data(uint8 *data, size_t size);
//	void recv_comm_data(uint8 *data, size_t size);
	/// @name misc
	//@{
	int parami[VM::ParamiUnknown];
	void *paramv[VM::ParamvUnknown];
	int get_parami(int id);
	void set_parami(int id, int val);
	void *get_paramv(int id);
	void set_paramv(int id, void *val);
	void update_params();
	//@}
#endif
	/// @name for debug
	//@{
	uint64 current_clock();
	//@}

	/// @name disp message methods
	//@{
	void send_log_message(int level, const _TCHAR *levelstr, const _TCHAR *msg);

	void out_msg(const _TCHAR* msg, bool set, int sec);
	void out_msg(const _TCHAR* msg, bool set = true);
	void out_info(const _TCHAR* msg, bool set, int sec);
	void out_info(const _TCHAR* msg, bool set = true);
	void out_infof(const _TCHAR* format, ...);

	MsgBoard *get_msgboard() { return msgboard; }
	//@}

	/// @name static functions
	//@{
	static void copy_initial_path(_TCHAR *dst, const _TCHAR *file_path);

	//@}
};

#endif
