/** @file fileio.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@brief [ file i/o ]
*/

#include "fileio.h"
#include <wchar.h>
#include "utility.h"

FILEIO::FILEIO()
{
	fp = NULL;
}

FILEIO::~FILEIO(void)
{
	Fclose();
}

bool FILEIO::IsProtected(const _TCHAR *filename)
{
#if defined(_WIN32)
	// for windows
	struct _stat fs;

	if (_tstat(filename, &fs)) {
		return false;
	}
	if ((fs.st_mode & _S_IWRITE) == 0) {
		return true;
	}
#else
	char filename_a[_MAX_PATH];
	UTILITY::cconv_from_native_path(filename, filename_a, _MAX_PATH);

	struct stat fs;

	if (stat(filename_a, &fs)) {
		return false;
	}
	if ((fs.st_mode & S_IWUSR) == 0) {
		return true;
	}
#if defined(__APPLE__) && defined(__MACH__)
	// locked file
	if ((fs.st_flags & 2) != 0) {
		return true;
	}
#endif
#endif

	return false;
}

bool FILEIO::Fopen(const _TCHAR *filename, int mode)
{
	Fclose();

#if defined(_WIN32) || !defined(_UNICODE)
	switch(mode) {
	case FILEIO_READ_BINARY:
		return ((fp = _tfopen(filename, _T("rb"))) == NULL) ? false : true;
	case FILEIO_WRITE_BINARY:
		return ((fp = _tfopen(filename, _T("wb"))) == NULL) ? false : true;
	case FILEIO_READ_WRITE_BINARY:
		return ((fp = _tfopen(filename, _T("r+b"))) == NULL) ? false : true;
	case FILEIO_READ_ASCII:
		return ((fp = _tfopen(filename, _T("r"))) == NULL) ? false : true;
	case FILEIO_WRITE_ASCII:
		return ((fp = _tfopen(filename, _T("w"))) == NULL) ? false : true;
	case FILEIO_READ_WRITE_ASCII:
		return ((fp = _tfopen(filename, _T("r+w"))) == NULL) ? false : true;
	}
#else
	char filename_a[_MAX_PATH];
	UTILITY::cconv_from_native_path(filename, filename_a, _MAX_PATH);
	switch(mode) {
	case FILEIO_READ_BINARY:
		return ((fp = fopen(filename_a, "rb")) == NULL) ? false : true;
	case FILEIO_WRITE_BINARY:
		return ((fp = fopen(filename_a, "wb")) == NULL) ? false : true;
	case FILEIO_READ_WRITE_BINARY:
		return ((fp = fopen(filename_a, "r+b")) == NULL) ? false : true;
	case FILEIO_READ_ASCII:
		return ((fp = fopen(filename_a, "r")) == NULL) ? false : true;
	case FILEIO_WRITE_ASCII:
		return ((fp = fopen(filename_a, "w")) == NULL) ? false : true;
	case FILEIO_READ_WRITE_ASCII:
		return ((fp = fopen(filename_a, "r+w")) == NULL) ? false : true;
	}
#endif
	return false;
}

void FILEIO::Fclose()
{
	if(fp) {
		fclose(fp);
	}
	fp = NULL;
}

int FILEIO::Fgetc()
{
	return fgetc(fp);
}

int FILEIO::Fputc(int c)
{
	return fputc(c, fp);
}

char *FILEIO::Fgets(char *buffer, int max_size)
{
	return fgets(buffer, max_size, fp);
}
wchar_t *FILEIO::Fgets(wchar_t *buffer, int max_size)
{
	return fgetws(buffer, max_size, fp);
}

size_t FILEIO::Fread(void* buffer, size_t size, size_t count)
{
	return fread(buffer, size, count, fp);
}

size_t FILEIO::Fwrite(void* buffer, size_t size, size_t count)
{
	return fwrite(buffer, size, count, fp);
}

uint32 FILEIO::Fseek(long offset, int origin)
{
	switch(origin) {
	case FILEIO_SEEK_CUR:
		return fseek(fp, offset, SEEK_CUR);
	case FILEIO_SEEK_END:
		return fseek(fp, offset, SEEK_END);
	case FILEIO_SEEK_SET:
		return fseek(fp, offset, SEEK_SET);
	}
	return 0xFFFFFFFF;
}

uint32 FILEIO::Ftell()
{
	return (uint32)ftell(fp);
}

void FILEIO::Remove(_TCHAR *filename)
{
#if defined(_WIN32)
	_tunlink(filename);
#else
	char filename_a[_MAX_PATH];
	UTILITY::cconv_from_native_path(filename, filename_a, _MAX_PATH);
	unlink(filename_a);
#endif
}
