/** @file utility.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2014.04.01 -

	@brief [ utility ]
*/

#include "utility.h"
#include <ctype.h>
#ifdef _UNICODE
#include "ConvertUTF.h"
#endif

#ifdef _WIN32
// #define USE_WINDOWS_API
#include <Windows.h>
#include <Shlwapi.h>

#define PATH_SEP_CHAR1	_T('\\')
#define PATH_SEP_STR1	_T("\\")
#define PATH_GPARENT_STR1	_T("...\\")
#define PATH_PARENT_STR1	_T("..\\")
#define PATH_CURRENT_STR1	_T(".\\")

#define PATH_SEP_CHAR2	_T('/')
#define PATH_SEP_STR2	_T("/")
#define PATH_GPARENT_STR2	_T(".../")
#define PATH_PARENT_STR2	_T("../")
#define PATH_CURRENT_STR2	_T("./")

#else	/* !_WIN32 */

#define PATH_SEP_CHAR1	_T('/')
#define PATH_SEP_STR1	_T("/")
#define PATH_GPARENT_STR1	_T(".../")
#define PATH_PARENT_STR1	_T("../")
#define PATH_CURRENT_STR1	_T("./")

#define PATH_SEP_CHAR2	_T('\\')
#define PATH_SEP_STR2	_T("\\")
#define PATH_GPARENT_STR2	_T("...\\")
#define PATH_PARENT_STR2	_T("..\\")
#define PATH_CURRENT_STR2	_T(".\\")

#endif

namespace UTILITY
{

// ----------------------------------------------------------------------
// file path conversion

/**
 *	match extension
 */
bool check_file_extension(const _TCHAR* file_path, const _TCHAR* ext)
{
	int nam_len = (int)_tcslen(file_path);
	int ext_len = (int)_tcslen(ext);

#ifdef _WIN32
	if(nam_len >= ext_len && _tcsnicmp(&file_path[nam_len - ext_len], ext, ext_len) == 0) {
#else
#ifdef _UNICODE
	if(nam_len >= ext_len && wcsncasecmp(&file_path[nam_len - ext_len], ext, ext_len) == 0) {
#else
	if(nam_len >= ext_len && strncasecmp(&file_path[nam_len - ext_len], ext, ext_len) == 0) {
#endif
#endif
		return true;
	}
	return false;
}

/**
 *	convert path separator
 */
void convert_path_separator(_TCHAR *path)
{
	size_t len = _tcslen(path);

	for(size_t i=0; i<len; i++) {
		if (path[i] == PATH_SEP_CHAR2) {
			path[i]=PATH_SEP_CHAR1;
		}
	}
}

/**
 *	add path separator
 */
void add_path_separator(_TCHAR *path, size_t maxlen)
{
	size_t len = _tcslen(path);

	if (len > 0 && len < (maxlen - 1)) {
		if (path[len-1] != PATH_SEP_CHAR1) {
			path[len]=PATH_SEP_CHAR1;
			path[len+1]=_T('\0');
		}
	}
}

/**
 *	get parent path
 */
void get_parent_dir(_TCHAR *path)
{
	int pt = (int)_tcslen(path);
	while(pt >= 0 && path[pt] != PATH_SEP_CHAR1) {
		pt--;
	}
	path[pt + 1] = _T('\0');
}

/**
 *	get ancestor path (end of path must be sep char)
 */
void get_ancestor_dir(_TCHAR *path, int ancestor_num)
{
	int pt = (int)_tcslen(path);
	int num = 0;

	ancestor_num++;
	while(pt > 0 && num < ancestor_num) {
		pt--;
		if (path[pt] == PATH_SEP_CHAR1) num++;
	}
	if (num == ancestor_num) path[pt + 1] = _T('\0');
}

/**
 *	get dirname and basename (pattern 1)
 */
void get_dir_and_basename(const _TCHAR *path, _TCHAR *dir, _TCHAR *name)
{
	const _TCHAR *p = NULL;

	if (dir != NULL)  {
		memset(dir,  0, _MAX_PATH * sizeof(_TCHAR));
	}
	if (name != NULL) {
		memset(name, 0, _MAX_PATH * sizeof(_TCHAR));
	}
	p = _tcsrchr(path, PATH_SEP_CHAR1);
	if (p != NULL) {
		if (dir  != NULL) {
			_tcsncpy(dir,  path, p-path+1);
			dir[p-path+1] = _T('\0');
		}
		if (name != NULL) {
			_tcsncpy(name, p+1, _MAX_PATH-1);
		}
	} else {
		if (name != NULL) {
			_tcsncpy(name, path, _MAX_PATH-1);
		}
	}
	return;
}

/**
 *	get dirname and basename (pattern 2)
 */
void get_dir_and_basename(_TCHAR *path, _TCHAR *name)
{
	_TCHAR *p = NULL;

	if (name != NULL) memset(name, 0, sizeof(_TCHAR) * _MAX_PATH);

	p = _tcsrchr(path, PATH_SEP_CHAR1);
	if (p != NULL) {
		if (name != NULL) _tcsncpy(name, p+1, _MAX_PATH-1);
		*(p+1) = _T('\0');
	} else {
		if (name != NULL) _tcsncpy(name, path, _MAX_PATH-1);
		*path = _T('\0');
	}
	return;
}

/**
 *	triming center of long string
 */
_TCHAR *trim_center(const _TCHAR *str, int maxlen)
{
	static _TCHAR buff[_MAX_PATH];

	int srclen = (int)_tcslen(str);
	memset(buff, 0, sizeof(buff));
	if (srclen > maxlen + 4) {
		int len = maxlen / 2 - 2;
#ifndef _UNICODE
		len += utf8firstpos(&str[len-1]);
#endif
		_tcsncpy(buff, str, len);
		_tcscat(buff, _T("..."));
		int pos = srclen - len;
#ifndef _UNICODE
		pos += utf8firstpos(&str[pos]);
#endif
		_tcscat(buff, &str[pos]);
	} else {
		_tcscpy(buff, str);
	}
	return buff;
}

/**
 *	make relative path if match base_path
 */
bool make_relative_path(const _TCHAR *base_path, _TCHAR *path)
{
	_TCHAR new_path[_MAX_PATH];

#ifdef USE_WINDOWS_API	// Use Windows API //
	if (!PathRelativePathTo(new_path, base_path, FILE_ATTRIBUTE_DIRECTORY, path, 0)) {
		return false;
	}
	_tcscpy(path, new_path);

#else	// ! Use Windows API //
	_TCHAR s_base_path[_MAX_PATH];
	_TCHAR s_path[_MAX_PATH];

	_tcscpy(s_base_path, slim_path(base_path));
	_tcscpy(s_path, slim_path(path));

	int len1 = (int)_tcslen(s_base_path);

	if (len1 == 0) {
		return true;
	}
	if (_tcsncmp(s_path, s_base_path, 2) != 0) {
		// if first 2 chars are different, stop convert it
		return true;
	}

	const _TCHAR *p1st = s_base_path;
	const _TCHAR *p1ed = NULL;
	_TCHAR *p2st = s_path;
	int parent = 0;

	while (p1st < s_base_path + len1 && (p1ed = _tcschr(p1st, PATH_SEP_CHAR1)) != NULL) {
		int len_part = (int)(p1ed - s_base_path + 1);
		if (_tcsncmp(s_path, s_base_path, len_part) == 0) {
			// same
			p2st = s_path + len_part;
		} else {
			// different
			parent++;
		}
		p1st = p1ed + 1;
	}

	// add relative path
	memset(new_path, 0, sizeof(new_path));
	for(; parent > 0; parent--) {
		_tcscat(new_path, PATH_PARENT_STR1);
	}
	_tcscat(new_path, p2st);

	_tcscpy(path, new_path);

#endif	// end of Use Windows API //
	return true;
}

/**
 *	make absolute path
 */
bool make_absolute_path(const _TCHAR *base_path, _TCHAR *path)
{
	int len1 = (int)_tcslen(base_path);

	if (len1 == 0) {
		return false;
	}

	_TCHAR new_path[_MAX_PATH];

#ifdef USE_WINDOWS_API	// Use Windows API //
	if (!PathCombine(new_path, base_path, path)) {
		return false;
	}
	_tcscpy(path, new_path);

#else	// ! Use Windows API //

#ifdef _WIN32
	// drive letter or root
	if ((_istalpha(path[0]) && path[1] == _T(':')) || path[0] == PATH_SEP_CHAR1) {
#else
	// root
	if (path[0] == PATH_SEP_CHAR1) {
#endif
		return true;
	}

	_tcscpy(new_path, slim_path(base_path));

	int parent = 0;
	_TCHAR *p2st = slim_path(path, &parent);

	// trim base path
	_TCHAR *p1st = NULL;
	for(; parent >= 0; parent--) {
		p1st = _tcsrchr(new_path, PATH_SEP_CHAR1);
		if (p1st == NULL) {
			break;
		}
		*p1st = _T('\0');
	}
	if (p1st == NULL) {
		// parent of root dir ???
		// so root
		p1st = new_path;
#ifdef _WIN32
		// if drive letter
		if (_istalpha(p1st[0]) && p1st[1] == _T(':')) {
			p1st += 2;
		}
#endif
	}
	*p1st = PATH_SEP_CHAR1;
	p1st++;
	*p1st = _T('\0');

	// add
	_tcscpy(p1st, p2st);
	_tcscpy(path, new_path);
#endif	// end of Use Windows API //
	return true;
}

/**
 *	get long full pathname
 */
void get_long_full_path_name(const _TCHAR* src, _TCHAR* dst)
{
	_TCHAR tmp[_MAX_PATH];

	memset(tmp, 0, sizeof(tmp));
#ifdef _WIN32
	if(GetFullPathName(src, _MAX_PATH, tmp, NULL) == 0) {
		// not restrict
		_tcsncpy(tmp, src, _MAX_PATH);
		_tcsncpy(dst, tmp, _MAX_PATH);
	}
	else if(GetLongPathName(tmp, dst, _MAX_PATH) == 0) {
		_tcsncpy(dst, tmp, _MAX_PATH);
	}
#else
	_tcsncpy(tmp, src, _MAX_PATH);
	_tcsncpy(dst, tmp, _MAX_PATH);
#endif
}

/**
 *	slim path
 */
_TCHAR *slim_path(const _TCHAR *path, int *parent_num)
{
	static _TCHAR new_path[_MAX_PATH];

	const _TCHAR *p2st = path;
	const _TCHAR *p2ed = path;
	const _TCHAR *p2sep = NULL;
	int len2 = (int)_tcslen(path);
	int parent = 0;
	struct poslen_st {
		int pos;
		int len;
	} child_stack[64];
	int child = 0;

	memset(new_path, 0, sizeof(new_path));

	// if root
	if (_tcsncmp(p2st, PATH_SEP_STR1, 1) == 0) {
		// push path
		if (child < 64) {
			child_stack[child].pos = 0;
			child_stack[child].len = 1;
			child++;
		}
		p2st++;
		p2ed = p2st;
	}

	// parse parent path
	while (p2st < path + len2) {
		if (_tcsncmp(p2st, PATH_GPARENT_STR1, 4) == 0) {
			child-=2;
			if (child < 0) {
				parent-=child;
				child=0;
			}
			p2st += 4;
			p2ed = p2st;
		} else if (_tcsncmp(p2st, PATH_PARENT_STR1, 3) == 0) {
			child--;
			if (child < 0) {
				parent-=child;
				child=0;
			}
			p2st += 3;
			p2ed = p2st;
		} else if (_tcsncmp(p2st, PATH_CURRENT_STR1, 2) == 0) {
			p2st += 2;
			p2ed = p2st;
		} else if (_tcsncmp(p2st, PATH_SEP_STR1, 1) == 0) {
			p2st += 1;
			p2ed = p2st;
		} else if ((p2sep = _tcschr(p2st, PATH_SEP_CHAR1)) != NULL) {
			if (p2st < p2sep) {
				// push path
				if (child < 64) {
					child_stack[child].pos = (int)(p2st - path);
					child_stack[child].len = (int)(p2sep - p2st + 1);
					child++;
				}
				p2st = p2sep + 1;
			} else {
				break;
			}
		} else {
			break;
		}
	}
	if (parent_num == NULL) {
		for(; parent > 0; parent--) {
			_tcsncat(new_path, PATH_PARENT_STR1, 3);
		}
	} else {
		*parent_num = parent;
	}
	for(int i=0; i<child; i++) {
		_tcsncat(new_path, path + child_stack[i].pos, child_stack[i].len);
	}
	_tcscat(new_path, p2st);

	return new_path;
}

// ----------------------------------------------------------------------
// string conversion

#ifdef _UNICODE
	/* _UNICODE */

bool conv_tchar_to_utf8(const wchar_t *srcStart, int srcMaxSize, char *dstStart, int dstMaxSize)
{
	UTF8  **dst_st = (UTF8  **)&dstStart;
	UTF8   *dst_ed = (UTF8  *)(dstStart + dstMaxSize);
	ConversionResult rc = conversionOK;
	if (sizeof(wchar_t) == 4) {
		UTF32 **src_st = (UTF32 **)&srcStart;
		UTF32  *src_ed = (UTF32 *)(srcStart + srcMaxSize);
		rc = ConvertUTF32toUTF8((const UTF32 **)src_st, (const UTF32 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else if (sizeof(wchar_t) == 2) {
		UTF16 **src_st = (UTF16 **)&srcStart;
		UTF16  *src_ed = (UTF16 *)(srcStart + srcMaxSize);
		rc = ConvertUTF16toUTF8((const UTF16 **)src_st, (const UTF16 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else {
		return false;
	}
	if (rc != conversionOK) {
		return false;
	}
	return true;
}

void wconv_from_native_path(const wchar_t *src, wchar_t *dst, int maxSize)
{
	memset(dst, 0, maxSize * sizeof(wchar_t));
	wcsncpy(dst, src, maxSize > _MAX_PATH ? _MAX_PATH : maxSize);
	return;
}

void cconv_from_native_path(const wchar_t *src, char *dst, int maxSize)
{
	memset(dst, 0, maxSize);

#ifdef _WIN32
	int rc = 0;

	// WideChar -> UTF-8
	rc = WideCharToMultiByte(CP_UTF8, 0, src, -1, dst, maxSize, NULL, NULL);
#else
	bool rc = false;
	int len = (int)wcslen(src);
	len = len > maxSize ? maxSize : len;

	// WideChar -> UTF-8
	rc = conv_tchar_to_utf8(src, len, dst, maxSize);
#endif
	return;
}

/// convert from wide char to UTF-8 char on the UNICODE environment
int wcs_to_mbs(char *dst, const wchar_t *src, int maxSize)
{
	memset(dst, 0, sizeof(char) * maxSize);
#ifdef _WIN32
	int len = 0;

	// WideChar -> UTF-8
	len = WideCharToMultiByte(CP_UTF8, 0, src, -1, dst, maxSize, NULL, NULL);
#else
	int len = (int)wcslen(src);
	len = len > maxSize ? maxSize : len;

	// WideChar -> UTF-8
	conv_tchar_to_utf8(src, len, dst, maxSize);
#endif
	len = (int)strlen(dst);
	return len;
}

bool conv_utf8_to_tchar(const char *srcStart, int srcMaxSize, wchar_t *dstStart, int dstMaxSize)
{
	UTF8  **src_st = (UTF8  **)&srcStart;
	UTF8   *src_ed = (UTF8  *)(srcStart + srcMaxSize);
	ConversionResult rc = conversionOK;
	if (sizeof(wchar_t) == 4) {
		UTF32 **dst_st = (UTF32 **)&dstStart;
		UTF32  *dst_ed = (UTF32 *)(dstStart + dstMaxSize);
		rc = ConvertUTF8toUTF32((const UTF8 **)src_st, (const UTF8 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else if (sizeof(wchar_t) == 2) {
		UTF16 **dst_st = (UTF16 **)&dstStart;
		UTF16  *dst_ed = (UTF16 *)(dstStart + dstMaxSize);
		rc = ConvertUTF8toUTF16((const UTF8 **)src_st, (const UTF8 *)src_ed
		, dst_st, dst_ed, lenientConversion);
	} else {
		return false;
	}
	if (rc != conversionOK) {
		return false;
	}
	return true;
}

void conv_to_native_path(const wchar_t *src, wchar_t *dst, int maxSize)
{
	memset(dst, 0, maxSize * sizeof(wchar_t));
	wcsncpy(dst, src, maxSize > _MAX_PATH ? _MAX_PATH : maxSize);
	// chomp lf
	size_t lastpos = wcslen(dst)-1;
	if (dst[lastpos-1] == L'\r' && dst[lastpos] == L'\n') {
		dst[lastpos-1] = L'\0';
	} else if (dst[lastpos] == L'\n' || dst[lastpos] == L'\r') {
		dst[lastpos] = L'\0';
	}
	return;
}

void conv_to_native_path(const char *src, wchar_t *dst, int maxSize)
{
#ifdef _WIN32
	int rc = 0;

	// UTF-8 -> WideChar
	rc = MultiByteToWideChar(CP_UTF8, 0, src, -1, dst, maxSize * sizeof(wchar_t));
#else
	bool rc = false;
	int len = (int)strlen(src);
	len = len > maxSize ? maxSize : len;

	// UTF-8 -> WideChar
	rc = conv_utf8_to_tchar(src, len, dst, maxSize);
#endif
	// chomp lf
	size_t lastpos = wcslen(dst)-1;
	if (dst[lastpos-1] == L'\r' && dst[lastpos] == L'\n') {
		dst[lastpos-1] = L'\0';
	} else if (dst[lastpos] == '\n' || dst[lastpos] == L'\r') {
		dst[lastpos] = L'\0';
	}
	return;
}

int mbs_to_wcs(wchar_t *dst, const char *src, int maxSize)
{
	memset(dst, 0, sizeof(wchar_t) * maxSize);
#ifdef _WIN32
	int len = 0;

	// MBS -> WideChar
	len = MultiByteToWideChar(CP_ACP, 0, src, -1, dst, maxSize);
#else
	int len = strlen(src);
	len = len > maxSize ? maxSize : len;

	// UTF-8 -> WideChar
	conv_utf8_to_tchar(src, len, dst, maxSize);
#endif
	// chomp lf
	size_t lastpos = wcslen(dst)-1;
	if (dst[lastpos-1] == L'\r' && dst[lastpos] == L'\n') {
		dst[lastpos-1] = L'\0';
	} else if (dst[lastpos] == L'\n' || dst[lastpos] == L'\r') {
		dst[lastpos] = L'\0';
	}
	len = (int)wcslen(dst);
	return len;
}

#else	/* !_UNICODE */

bool conv_mbs_to_utf8(const char *srcStart, int srcMaxSize, char *dstStart, int dstMaxSize)
{
#ifdef _WIN32
	int rc = 0;
	wchar_t dstTemp[1024];
	// MBS -> WideChar
	rc = MultiByteToWideChar(CP_ACP, 0, srcStart, -1, dstTemp, 1024);
	if (rc > 0) {
		// WideChar -> UTF-8
		rc = WideCharToMultiByte(CP_UTF8, 0, dstTemp, -1, dstStart, dstMaxSize, NULL, NULL);
	}
	if (rc == 0) {
		// cannot convert
		if (dstStart != srcStart) {
			memset(dstStart, 0, dstMaxSize);
			strncpy(dstStart, srcStart, dstMaxSize);
		}
		return false;
	}
#else
	if (dstStart != srcStart) {
		memset(dstStart, 0, dstMaxSize);
		strncpy(dstStart, srcStart, dstMaxSize);
	}
#endif
	return true;
}

void cconv_from_native_path(const char *src, char *dst, int maxSize)
{
	conv_mbs_to_utf8(src, _MAX_PATH, dst, maxSize);
	return;
}

/// copy chars on the ASCII(MBCS) environment
int wcs_to_mbs(char *dst, const char *src, int maxSize)
{
	memset(dst, 0, sizeof(char) * maxSize);

	strncpy(dst, src, maxSize);

	int len = (int)strlen(dst);
	return len;
}

bool conv_utf8_to_mbs(const char *srcStart, int srcMaxSize, char *dstStart, int dstMaxSize)
{
#ifdef _WIN32
	int rc = 0;
	wchar_t dstTemp[1024];
	// UTF-8 -> WideChar
	rc = MultiByteToWideChar(CP_UTF8, 0, srcStart, -1, dstTemp, 1024);
	if (rc > 0) {
		// WideChar -> MBS
		rc = WideCharToMultiByte(CP_ACP, 0, dstTemp, -1, dstStart, dstMaxSize, NULL, NULL);
	}
	if (rc == 0) {
		// cannot convert
		if (srcStart != dstStart) {
			memset(dstStart, 0, dstMaxSize);
			strncpy(dstStart, srcStart, dstMaxSize);
		}
		return false;
	}
#else
	if (srcStart != dstStart) {
		memset(dstStart, 0, dstMaxSize);
		strncpy(dstStart, srcStart, dstMaxSize);
	}
#endif
	return true;
}

void conv_to_native_path(const char *src, char *dst, int maxSize)
{
	conv_utf8_to_mbs(src, _MAX_PATH, dst, maxSize);

	// chomp lf
	size_t lastpos = strlen(dst)-1;
	if (dst[lastpos-1] == '\r' && dst[lastpos] == '\n') {
		dst[lastpos-1] = '\0';
	} else if (dst[lastpos] == '\n' || dst[lastpos] == '\r') {
		dst[lastpos] = '\0';
	}
	return;
}

int mbs_to_wcs(char *dst, const char *src, int maxSize)
{
	memset(dst, 0, sizeof(char) * maxSize);

	strncpy(dst, src, maxSize);

	// chomp lf
	size_t lastpos = strlen(dst)-1;
	if (dst[lastpos-1] == '\r' && dst[lastpos] == '\n') {
		dst[lastpos-1] = '\0';
	} else if (dst[lastpos] == '\n' || dst[lastpos] == '\r') {
		dst[lastpos] = '\0';
	}
	int len = (int)strlen(dst);
	return len;
}

#endif	/* _UNICODE */


// ----------------------------------------------------------------------
// string format

#ifdef _UNICODE
const wchar_t *concat(const wchar_t *src, ...) {
	va_list ap;
	va_start(ap, src);
	const wchar_t *p = concatv(src, ap);
	va_end(ap);
	return p;
}
const wchar_t *concatv(const wchar_t *src, va_list ap) {
	static wchar_t str[1024];
	const wchar_t *src2;
	wcscpy(str, src);
	while((src2 = va_arg(ap, const wchar_t *)) != NULL) {
		wcscat(str, src2);
	}
	return str;
}
void concat(wchar_t *dst, size_t count, const wchar_t *src, ...) {
	va_list ap;
	va_start(ap, src);
	const wchar_t *p = concatv(src, ap);
	va_end(ap);
	wcsncpy(dst, p, count);
}
#endif /* _UNICODE */

const char *concat(const char *src, ...) {
	va_list ap;
	va_start(ap, src);
	const char *p = concatv(src, ap);
	va_end(ap);
	return p;
}
const char *concatv(const char *src, va_list ap) {
	static char str[1024];
	const char *src2;
	strcpy(str, src);
	while((src2 = va_arg(ap, const char *)) != NULL) {
		strcat(str, src2);
	}
	return str;
}
void concat(char *dst, size_t count, const char *src, ...) {
	va_list ap;
	va_start(ap, src);
	const char *p = concatv(src, ap);
	va_end(ap);
	strncpy(dst, p, count);
}


int stprintf(_TCHAR *dest, size_t count, const _TCHAR *format, ...) {
	va_list ap;
	va_start(ap, format);
	int len = UTILITY::vstprintf(dest, count, format, ap);
	va_end(ap);
	return len;
}

int vstprintf(_TCHAR *dest, size_t count, const _TCHAR *format, va_list ap) {
#ifdef _UNICODE
	_TCHAR wformat[1024];
	int wpos = 0;
	int len = (int)wcslen(format);
	wformat[wpos]=_T('\0');
	for(int pos=0; pos < len && wpos < 1022; ) {
		if (wcsncmp(&format[pos], _T("%s"), 2) == 0) {
			wcscpy(&wformat[wpos], _T("%ls"));
			pos+=2;
			wpos+=3;
		} else {
			wformat[wpos]=format[pos];
			pos++;
			wpos++;
		}
	}
	wformat[wpos]=_T('\0');
	return vswprintf(dest, count, wformat, ap);
#else
	return vsprintf(dest, format, ap);
#endif
}


int sprintf_utf8(char *dest, size_t count, const char *format, ...) {
	va_list ap;
	va_start(ap, format);
	int len = UTILITY::vsprintf_utf8(dest, count, format, ap);
	va_end(ap);
	return len;
}

#if (defined(_WIN32) && !defined(__MINGW32__))

/// Char code in format is UTF-8. So, convert to wide char and format a string using vwprintf.
/// Because a program sometimes occur exception in printf.

/// @param[out] dest
/// @param[in] count buffer size on dest
/// @param[in] format UTF-8 based printf style string
/// @param[in] ap
int vsprintf_utf8(char *dest, size_t count, const char *format, va_list ap) {
	wchar_t wformat[1024];
	wchar_t wdest[1024];
	char nformat[1024];
	char ndest[1024];
	int len;
	len = ::MultiByteToWideChar(CP_UTF8, 0, format, -1, wformat, 1024);
	if (len <= 0) return 0;
	len = ::WideCharToMultiByte(CP_ACP, 0, wformat, -1, nformat, 1024, NULL, NULL);
	if (len <= 0) return 0;
	len = vsprintf(ndest, nformat, ap);
	if (len <= 0) return 0;
	len = ::MultiByteToWideChar(CP_ACP, 0, ndest, -1, wdest, 1024);
	if (len <= 0) return 0;
	len = ::WideCharToMultiByte(CP_UTF8, 0, wdest, -1, dest, (int)count, NULL, NULL);
	return len;
}

#else	/* !_WIN32 && !__MINGW32__ */

int vsprintf_utf8(char *dest, size_t count, const char *format, va_list ap) {
	return vsprintf(dest, format, ap);
}

#endif	/* _WIN32 && !__MINGW32__ */

/// irresponsible
int utf8firstpos(const char *str)
{
	const char *p = str;
	int pos = 0;
	for(; *p != '\0'; p++, pos++) {
		if (((*p) & 0xc0) != 0x80) break;
	}
	return pos;
}

/// substring in last circle bracket
/// @return length of substring  -1:not found
int substr_in_bracket(const _TCHAR *src, _TCHAR *dst, int maxSize)
{
	const _TCHAR *ps,*pe;
	int len = 0;

	pe = _tcsrchr(src, _T(')'));
	if (pe == NULL) return -1;
	ps = _tcsrchr(src, _T('('));
	if (ps == NULL) return -1;
	len = (int)(pe - ps - 1);
	if (len < 0) return -1;
	if (len >= maxSize-1) len = maxSize-1;

	_tcsncpy(dst, ps+1, len);
	dst[len]=_T('\0');

	return len;
}

/// @brief trimming string at right size
/// @param[in,out] str
/// @param[in] sep trimming char or NULL
void rtrim(_TCHAR *str, const _TCHAR *sep)
{
	// trim white space
	for (int pos = (int)_tcslen(str) - 1; pos >= 0; pos--) {
		if (str[pos] == _T(' ') || str[pos] == _T('\t')
		|| (sep != NULL && *str != _T('\0') && _tcschr(sep, *str) != NULL)) {
			str[pos]=_T('\0');
		} else {
			break;
		}
	}
}

/// @brief get token from string that is separated by char
/// @param[in] str
/// @param[in] start_pos
/// @param[in] max_len
/// @param[out] word
/// @param[in] word_max_len
/// @param[in] sep separated char
/// @return >=0: next position / -1:last word
int get_token(const _TCHAR *str, int start_pos, int max_len, _TCHAR *word, int word_max_len, const _TCHAR sep)
{
	int pos = start_pos;

	word[0]=_T('\0');

	if (pos < 0) {
		return pos;
	}

	// ltrim
	for(; str[pos] != _T('\0') && pos < max_len; pos++) {
		if (str[pos] == _T(' ') || str[pos] == _T('\t')) {
			start_pos++;
		} else {
			break;
		}
	}
	// search next separator
	int len = 0;
	bool quote = false;
	for(; str[pos] != _T('\0') && pos < max_len && len < (word_max_len - 1); pos++) {
		if (str[pos] == _T('"')) {
			quote = !quote;
			continue;
		}
		if (quote == false && str[pos] == sep) {
			word[len]=_T('\0');
			break;
		}
		word[len++]=str[pos];
	}
	// end of line ?
	if (str[pos] == _T('\0') || pos == max_len || len == (word_max_len - 1)) {
		word[len]=_T('\0');
		pos = -1;
	} else {
		pos++;
	}
	// rtrim
	rtrim(word);

	return pos;
}

}	// end of UTILITY
