/** @file rec_video.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.05.18 -

	@brief [ record video ]
*/

#include "rec_video.h"
#include "../emu.h"
#include "../sdl_utils.h"

#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_VFW
#include "vfw/vfw_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_QTKIT
#include "qtkit/qt_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_FFMPEG
#include "ffmpeg/ffm_rec_video.h"
#endif
#endif /* USE_REC_VIDEO */

REC_VIDEO::REC_VIDEO(EMU *new_emu)
{
	emu = new_emu;
	now_recording = false;
	rec_type = 0;
	rec_fps = 0;
	rec_count = 0;
	rec_denomi = 1;
	memset(rec_path, 0, sizeof(rec_path));
	rec_surface = NULL;

#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_VFW
	vfwvideo = new VFW_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_QTKIT
	qtvideo = new QT_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	ffmvideo = new FFM_REC_VIDEO(new_emu, this);
#endif
#endif /* USE_REC_VIDEO */
}

REC_VIDEO::~REC_VIDEO()
{
#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_FFMPEG
	delete ffmvideo;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	delete qtvideo;
#endif
#ifdef USE_REC_VIDEO_VFW
	delete vfwvideo;
#endif
#endif /* USE_REC_VIDEO */

	SDL_UTILS::release_surface(rec_surface);
}

bool REC_VIDEO::IsEnabled(int type)
{
	bool rc = false;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		rc = true;
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		rc = qtvideo->IsEnabled();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		rc = ffmvideo->IsEnabled();
		break;
#endif
	default:
		break;
	}
#endif
	return rc;
}

const _TCHAR **REC_VIDEO::GetCodecList(int type)
{
	const _TCHAR **list = NULL;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		list = qtvideo->GetCodecList();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetCodecList();
		break;
#endif
	default:
		break;
	}
#endif
	return list;
}

const _TCHAR **REC_VIDEO::GetQualityList(int type)
{
	const _TCHAR **list = NULL;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		list = qtvideo->GetQualityList();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetQualityList();
		break;
#endif
	default:
		break;
	}
#endif
	return list;
}

bool REC_VIDEO::Start(int type, int fps, SDL_Rect *srcrect, SDL_Surface *srcsurface, bool show_dialog)
{
#ifdef USE_REC_VIDEO
	if (type > 0) {
		rec_type = type;
	}
	if(fps > 0) {
		rec_fps = fps;
		rec_count = 0;
		rec_denomi = 60 / fps;
	}

	if (rec_type <= 0 || rec_fps <= 0) {
		return false;
	}

	CreateFileName(rec_path);

	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		rec_surface = SDL_UTILS::release_surface(rec_surface);
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 2);
		if (rec_surface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create surface for recording video: %s"), SDL_GetError());
			return false;
		}
		now_recording = vfwvideo->Start(rec_path, rec_fps, srcrect, rec_surface, show_dialog);
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		rec_surface = SDL_UTILS::release_surface(rec_surface);
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 1);
		if (rec_surface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create surface for recording video: %s"), SDL_GetError());
			return false;
		}
		char crec_path[_MAX_PATH];
		UTILITY::wcs_to_mbs(crec_path, rec_path);
		now_recording = qtvideo->Start(crec_path, rec_fps, rec_surface, show_dialog);
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		// height set minus, so create the flipped surface
		rec_surface = SDL_UTILS::release_surface(rec_surface);
#if defined(__APPLE__) && defined(__MACH__)
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 1);
#else
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 1);
#endif
		if (rec_surface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create surface for recording video: %s"), SDL_GetError());
			return false;
		}
		now_recording = ffmvideo->Start(rec_path, rec_fps, srcrect, rec_surface, show_dialog);
		break;
#endif
	default:
		break;
	}
#endif
	return now_recording;
}

void REC_VIDEO::Stop()
{
	if (!now_recording) return;
#ifdef USE_REC_VIDEO
	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		vfwvideo->Stop();
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		qtvideo->Stop();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		ffmvideo->Stop();
		break;
#endif
	default:
		break;
	}
	rec_surface = SDL_UTILS::release_surface(rec_surface);
#endif
	now_recording = false;
}

bool REC_VIDEO::Restart()
{
	if (!now_recording) return false;
#ifdef USE_REC_VIDEO
	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		now_recording = vfwvideo->Restart();
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		now_recording = qtvideo->Restart();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		now_recording = ffmvideo->Restart();
		break;
#endif
	default:
		break;
	}
#endif
	return now_recording;
}

bool REC_VIDEO::Record(SDL_Rect *srcrect, SDL_Surface *srcsurface, SDL_Rect *dstrect)
{
	if (!now_recording) return false;

#ifdef USE_REC_VIDEO
	// copy surface
	SDL_Rect re = { 0, 0, dstrect->w, dstrect->h };

	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
			SDL_UTILS::copy_surface_flipped(srcsurface, srcrect, rec_surface, &re);
		} else {
			SDL_UTILS::zoomin_surface_flipped(srcsurface, srcrect, rec_surface, &re);
		}
//		emu->mix_ledbox(rec_surface);
		now_recording = vfwvideo->Record(rec_surface);
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
#ifndef USE_SDL2
			SDL_UTILS::copy_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitSurface(srcsurface, srcrect, rec_surface, &re);
#endif
		} else {
#ifndef USE_SDL2
			SDL_UTILS::zoomin_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitScaled(srcsurface, srcrect, rec_surface, &re);
#endif
		}
//		emu->mix_ledbox(rec_surface);
		now_recording = qtvideo->Record(rec_surface);
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
#ifndef USE_SDL2
			SDL_UTILS::copy_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitSurface(srcsurface, srcrect, rec_surface, &re);
#endif
		} else {
#ifndef USE_SDL2
			SDL_UTILS::zoomin_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitScaled(srcsurface, srcrect, rec_surface, &re);
#endif
		}
//		emu->mix_ledbox(rec_surface);
		now_recording = ffmvideo->Record(rec_surface);
		break;
#endif
	default:
		break;
	}
#endif
	return now_recording;
}

bool REC_VIDEO::IsRecordFrame()
{
	if (!now_recording) return false;

	bool rc = (rec_count == 0);
	rec_count = ((rec_count + 1) % rec_denomi);
	return rc;
}

bool REC_VIDEO::Capture(int type, SDL_Rect *srcrect, SDL_Surface *srcsurface, SDL_Rect *dstrect)
{
	if (srcrect == NULL || srcsurface == NULL) {
		return false;
	}

	_TCHAR file_path[_MAX_PATH];
	CreateFileName(file_path);

	switch(type) {
	case CAPTURE_SCREEN_TYPE_BMP:
	{
		// add ext
		_tcscat(file_path, _T(".bmp"));

		// notice that format is same as screen because SaveBMP cannot out pixeldata
		SDL_Surface *bmpsurface = SDL_UTILS::create_surface(dstrect, srcsurface);
		if (bmpsurface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create screen to capture: %s"), SDL_GetError());
			return false;
		}

		// copy surface
		SDL_Rect re = { 0, 0, dstrect->w, dstrect->h };

		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
#ifndef USE_SDL2
			SDL_UTILS::copy_surface(srcsurface, srcrect, bmpsurface, &re);
#else
			SDL_BlitSurface(srcsurface, srcrect, bmpsurface, &re);
#endif
		} else {
#ifndef USE_SDL2
			SDL_UTILS::zoomin_surface(srcsurface, srcrect, bmpsurface, &re);
#else
			SDL_BlitScaled(srcsurface, srcrect, bmpsurface, &re);
#endif
		}

		// save surface
		char cfile_path[_MAX_PATH];
		UTILITY::wcs_to_mbs(cfile_path, file_path);
		if (SDL_SaveBMP(bmpsurface, cfile_path) < 0) {
			emu->out_logf(LOG_ERROR, _T("Cannot save screen captured file: %s"), SDL_GetError());
		}

		SDL_UTILS::release_surface(bmpsurface);

		break;
	}
	default:
		emu->out_log(LOG_ERROR,_T("Capture screen failed. Unknown type."));
		return false;
	}
	emu->out_log(LOG_INFO,(CT)_("Screen was saved successfully."));
	return true;
}

void REC_VIDEO::CreateFileName(_TCHAR *file_path)
{
	int tim[8];
	const _TCHAR *app_path;

	app_path = _tcslen(config.snapshot_path) > 0 ? config.snapshot_path : emu->application_path();
	emu->get_timer(tim);
	_stprintf(file_path, _T("%s%04d-%02d-%02d_%02d-%02d-%02d"), app_path, tim[0], tim[1], tim[2], tim[4], tim[5], tim[6]);
}
