/** @file memory.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ memory ]
*/

#include <time.h>
#include "memory.h"
#include "../../fileio.h"

//#define _DEBUG_RAM
//#define _DEBUG_CRAM

#define TRACE_PC_IDLE		0xffffffff
#define TRACE_PC_FETCH		0xfffffffe
#define TRACE_PC_WAITING	0xfffffffd
#define TRACE_PC_FIRE		0xfffffffc
#define TRACE_PC_FIRED		0xfffffffb

void MEMORY::SET_BANK(int s, int e, uint8* w, uint8* r) {
	int sb = (s) >> BANK_SIZE, eb = (e) >> BANK_SIZE;
	for(int i = sb; i <= eb; i++) {
		if((w) == wdmy) {
			wbank[i] = wdmy;
		} else if ((w) != NULL) {
			wbank[i] = (w) + (1 << BANK_SIZE) * (i - sb);
		}
		if((r) == rdmy) {
			rbank[i] = rdmy;
		} else if ((r) != NULL) {
			rbank[i] = (r) + (1 << BANK_SIZE) * (i - sb);
		}
	}
}

void MEMORY::initialize()
{
	rom_loaded = false;
	rom1802_loaded = false;
	rom1805_loaded = false;
	rom_loaded_at_first = false;

	memset(rom, 0xff, sizeof(rom));
	memset(rdmy, 0xff, sizeof(rdmy));
	memset(rom1802, 0xff, sizeof(rom1802));
	memset(rom1805, 0xff, sizeof(rom1805));

	// read rom image from file
	load_rom_files();

	// init memory map
	reset_membank();

	// clear trace counter
	trace_counter_id = -1;

	update_config();
}

void MEMORY::reset()
{
	// read rom image from file
	load_rom_files();

	memset(ram , 0, sizeof(ram));
	memset(ram2, 0, sizeof(ram2));

	memset(&ram[0xff00], 0xff, 16 * 15);

	// clear color ram
	srand((unsigned int)time(NULL));
	int co = (rand() % 3) * 3 + 1;	// 1 4 7
	int cg = (rand() % 2) * 24;		// 0 24
	memset(color_ram, 0, sizeof(color_ram));
	color_reg = co + (co == 7 ? cg : 0);

	// clear ig ram
	reset_igram();

	// init memory map
	reset_membank();

	// stop trace counter
	reset_tracecounter();

#ifdef USE_RTC
	rtc_reg_sel = 0;
	rtc_buf = 0;
#endif
}

void MEMORY::load_rom_files()
{
	// load rom
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	FILEIO* fio = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		if (!rom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sL3BAS.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(rom, 0x6000, 1);
				fio->Fclose();
				rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("L3BAS.ROM was loaded."));
			}
		}
		if (!rom_loaded) {
			// try to load another main rom.
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sROM1.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(rom, 0x5f00, 1);
				fio->Fclose();
				emu->out_log(LOG_INFO, (CT)_("ROM1.ROM was loaded."));
			}
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sROM2.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(&rom[0x5ff0], 16, 1);
				fio->Fclose();
				rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("ROM2.ROM was loaded."));
			}
		}
		if (!rom1802_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sMP1802.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(rom1802, 0x800, 1);
				fio->Fclose();
				rom1802_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("MP1802.ROM was loaded."));
			}
		}
		if (!rom1805_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sMP1805.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(rom1805, 0x800, 1);
				fio->Fclose();
				rom1805_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("MP1805.ROM was loaded."));
			}
		}
	}
	delete fio;

	if (!rom_loaded_at_first) {
		if (!rom_loaded) {
			emu->out_log(LOG_WARN, (CT)_("L3BAS.ROM couldn't be loaded."));
			emu->out_log(LOG_WARN, (CT)_("ROM1.ROM couldn't be loaded."));
			emu->out_log(LOG_WARN, (CT)_("ROM2.ROM couldn't be loaded."));
		}
		if (!rom1802_loaded) {
				emu->out_log(LOG_WARN, (CT)_("MP1802.ROM couldn't be loaded."));
		}
		if (!rom1805_loaded) {
				emu->out_log(LOG_WARN, (CT)_("MP1805.ROM couldn't be loaded."));
		}
		rom_loaded_at_first = true;
	}
}

void MEMORY::reset_igram()
{
	// clear ig ram data
	for(int co=0; co<3; co++) {
		for(int ch=0; ch<256; ch++) {
			for(int i=0; i<8; i+=2) {
				ig_ram[(co*256+ch)*8+i]  =0x00;
				ig_ram[(co*256+ch)*8+i+1]=0xff;
				if ((ch % 3) == co) {
					switch(i) {
						case 0:
							if ((ch >> 4) < 8) ig_ram[(co*256+ch)*8+i]=(1 << (ch >> 4));
							break;
						case 2:
							if ((ch >> 4) < 16) ig_ram[(co*256+ch)*8+i]=(1 << ((ch >> 4) - 8));
							break;
						case 4:
							if ((ch & 15) < 8) ig_ram[(co*256+ch)*8+i]=(1 << (ch & 15));
							break;
						case 6:
							if ((ch & 15) < 16) ig_ram[(co*256+ch)*8+i]=(1 << ((ch & 15) - 8));
							break;
					}
				}
			}
		}
	}
	ig_enable = 0;
	ig_enreg = 0;
}

void MEMORY::reset_membank()
{
	// init memory map
	mem_bank_reg1 = 0x0e;
	mem_bank_reg2 = 0;
//	vram_sel = 0;
	change_memory_bank();
}

void MEMORY::reset_tracecounter()
{
	// stop trace counter
	if (trace_counter_id != -1) {
		cancel_event(trace_counter_id);
	}
	trace_counter_id = -1;
	d_board->write_signal(SIG_CPU_NMI, 0, SIG_NMI_TRACE_MASK);
	pc = 0;
	pc_prev = TRACE_PC_IDLE;
}

void MEMORY::set_igmode(uint32 data)
{
	ig_enable = data & 0x01;
	d_disp->write_io8(0xffe9, data);
}

#ifdef USE_CPU_REAL_MACHINE_CYCLE
void MEMORY::write_data8w(uint32 addr, uint32 data, int *wait)
#else
void MEMORY::write_data8(uint32 addr, uint32 data)
#endif
{
	addr &= 0xffff;

#ifdef _DEBUG_RAM
	if (addr >= 0xff70 && addr <= 0xff7f) {
		uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
		emu->out_debugf("mw %04x=%02x %c",addr,data,ch);
	}
#endif

	// vram area
	if (ADDR_VRAM_START <= addr && addr < ADDR_VRAM_END) {
#ifdef _DEBUG_CRAM
//		if ((addr - ADDR_VRAM_START) < 0x0400) {
			uint8 ch=data;
			ch = (ch < 0x20 || ch > 0x7f) ? 0x20 : ch;
			emu->out_debugf("mw %04x=%02x->%02x %c c%02x->%02x",addr,ram[addr],data,ch,color_reg,color_ram[addr - ADDR_VRAM_START]);
//		}
#endif
		color_ram[addr - ADDR_VRAM_START] = (color_reg & 0x3f);
	}

	// ig_area
	if (ig_enable) {
		if (ADDR_IGRAM_START <= addr && addr < ADDR_IGRAM_END) {
			for(int i=0; i<3; i++) {
				if (ig_enreg & (1 << i)) {
					ig_ram[addr-ADDR_IGRAM_START+(ADDR_IGRAM_END-ADDR_IGRAM_START)*i] = data & 0xff;
				}
			}
			return;
		}
	}

	if (0xff00 <= addr && addr <= 0xffef) {
		ram[addr] = data;
	} else {
		wbank[addr >> BANK_SIZE][addr & ((1 << BANK_SIZE) - 1)] = data;
	}

	// memory mapped i/o
	switch(addr & 0xffff) {
#ifdef USE_FD1
		case 0xff00:
		case 0xff01:
		case 0xff02:
		case 0xff03:
			// fdc
			if (IOPORT_USE_5FDD) {
				d_fdc5->write_io8(addr - 0xff00, data);

//				emu->out_debugf("fdc5w a:%04x d:%02x",addr,data);
			}
			break;
		case 0xff04:
			// fdd drive select
			if (IOPORT_USE_5FDD) {
				d_fdd->write_io8(0, data);

//				emu->out_debugf(_T("fdd5w a:%04x d:%02x"),addr,data);
			}
			break;

		case 0xff18:
		case 0xff19:
		case 0xff1a:
		case 0xff1b:
		case 0xff1c:
		case 0xff1d:
		case 0xff1e:
		case 0xff1f:
			// fdc
			if (IOPORT_USE_3FDD) {
				d_fdc3->write_io8(addr - 0xff18, data);

//				emu->out_debugf("fdc3w a:%04x d:%02x",addr,data);
			}
			break;
		case 0xff20:
			// fdd drive select
			if (IOPORT_USE_3FDD) {
				d_fdd->write_io8(0, data);

//				emu->out_debugf("fdd3w a:%04x d:%02x",addr,data);
			}
			break;
#endif
		case 0xff30:
		case 0xff31:
		case 0xff32:
		case 0xff33:
			// psg
			if (IOPORT_USE_PSG6) {
				d_psg->write_io8(addr - 0xff30, data);
//				emu->out_debugf("psgw a:%04x d:%02x",addr,data);
			}
			break;
#ifdef USE_RTC
		case 0xff38:
			// rtc register select
			rtc_reg_sel = data & 0xff;
			break;

		case 0xff39:
			// rtc write
			if ((data & 0x80) == 0) {
				rtc_buf = data & 0xff;
			}
			break;
#endif
		case 0xff3c:
		case 0xff3d:
		case 0xff3e:
		case 0xff3f:
			// extended LPT board
			if (IOPORT_USE_EXPIA) {
				d_pia_ex->write_io8(addr - 0xff3c, data);
			}
			break;

		case 0xff40:
			// extended COM board
			if (IOPORT_USE_EXACIA) {
				d_comm1->write_io8(addr - 0xff40, data);
			}
			// through
		case 0xff41:
			// extended COM board
			if (IOPORT_USE_EXACIA) {
				d_acia_ex->write_io8(addr - 0xff40, data);
			}
			break;

		case 0xff70:
		case 0xff71:
		case 0xff72:
		case 0xff73:
		case 0xff74:
		case 0xff77:
		case 0xff78:
		case 0xff79:
		case 0xff7a:
		case 0xff7b:
		case 0xff7c:
		case 0xff7d:
		case 0xff7e:
		case 0xff7f:
			// psg9
			if (IOPORT_USE_PSG9) {
				d_psg9->write_io8(addr - 0xff70, data);
//				emu->out_debugf("psg9w a:%04x d:%02x",addr,data);
			}
			break;

		case 0xff75:
		case 0xff76:
			// psg9
			if (IOPORT_USE_PSG9) {
				d_psg9->write_io8(addr - 0xff70, data);
//				emu->out_debugf("psg9w a:%04x d:%02x",addr,data);
			}
			// kanji rom
			if (IOPORT_USE_KANJI) {
				d_kanji->write_io8(addr - 0xff75, data);
//				emu->out_debugf("kanjiw a:%04x d:%02x",addr,data);
			}
			break;

		case 0xffc0:
		case 0xffc1:
		case 0xffc2:
		case 0xffc3:
			// pia
			d_pia->write_io8(addr - 0xffc0, data);
			break;

		case 0xffc4:
			// acia
			d_comm->write_io8(addr - 0xffc4, data);
			// pass through
		case 0xffc5:
			// acia
			d_acia->write_io8(addr - 0xffc4, data);
			break;

		case 0xffc6:
		case 0xffc7:
			// crtc
			d_crtc->write_io8(addr - 0xffc6, data);
			break;

		case 0xffc8:
			// keyborad nmi (cannot write)
			break;
		case 0xffca:
			// timer irq
			break;
		case 0xffcb:
			// light pen flag (cannot write)
			break;
		case 0xffd0:
			// mode select
			d_disp->write_io8(addr, data);
			d_comm->write_signal(SIG_COMM_RS, data, 0x20);
			d_crtc->write_signal(SIG_HD46505_CHR_CLOCKS, CPU_CLOCKS * ((data & 0x80) ? 2 : 1), 0xffffffff);
//			mode_sel = data;
			break;
		case 0xffd1:
			// trace counter
			fetch_trace_counter();
			break;
		case 0xffd2:
			// remote switch
			d_cmt->write_signal(SIG_CMT_REMOTE, data, 0x80);
			break;
		case 0xffd3:
			// music sel
//			emu->out_debugf("ms %02x",data);
			d_sound->write_signal(SIG_SOUND_ON, 1, 1);
			d_sound->write_signal(SIG_SOUND_SIGNAL, data, 0x80);
			break;
		case 0xffd4:
			// time mask
			d_timer->write_io8(addr, data & 0x80);
			break;
		case 0xffd5:
			// light pen bl
//			emu->out_debugf(_T("lpengl: %02x"),data);
			d_key->write_io8(addr, data & 0x80);
			break;
		case 0xffd6:
			// interace sel
			d_disp->write_io8(addr, data & 0x08);
			break;
		case 0xffd7:
			// baud sel
			d_cmt->write_signal(SIG_CMT_BAUD_SEL, data, 0x01);
			break;
		case 0xffd8:
			// color register
#ifdef _DEBUG_CRAM
			emu->out_debugf("cw c%02x->%02x",color_reg,data);
#endif
			color_reg = data & 0xbf;
			break;
		case 0xffe0:
			// keyboard mode register
			d_key -> write_io8(addr, data);
			break;
		case 0xffe8:
			// bank register
			mem_bank_reg2 = data & 0xff;
			change_memory_bank();
			break;
		case 0xffe9:
			// IG mode register
			set_igmode(0x100 | data);
			break;
		case 0xffea:
			// IG en register
			ig_enreg = data & 0x07;
			break;
	}

}

#ifdef USE_CPU_REAL_MACHINE_CYCLE
uint32 MEMORY::read_data8w(uint32 addr, int *wait)
#else
uint32 MEMORY::read_data8(uint32 addr)
#endif
{
	uint32 data;
	addr &= 0xffff;

	if (pc_prev != TRACE_PC_IDLE) {
		if (pc_prev == TRACE_PC_FETCH) {
			// trace counter start
			set_trace_counter();
		} else if (pc_prev == TRACE_PC_FIRE) {
			pc_prev = TRACE_PC_FIRED;
		} else if (pc_prev == TRACE_PC_FIRED) {
			// trace counter reset
			clear_tracecounter();
		}
	}

	// vram area
	if (ADDR_VRAM_START <= addr && addr < ADDR_VRAM_END) {
#ifdef _DEBUG_CRAM
//		if ((addr - ADDR_VRAM_START) < 0x0400) {
			uint8 ch=ram[addr];
			ch = (ch < 0x20 || ch > 0x7f) ? 0x20 : ch;
			emu->out_debugf("mr %04x=%02x %c c%02x r%02x",addr,ram[addr],ch,color_ram[addr - ADDR_VRAM_START],color_reg);
//		}
#endif
		// 6bit
		// if MK bit is set, do not read from color ram.
		if ((color_reg & 0x80) == 0) {
			color_reg = (color_ram[addr - ADDR_VRAM_START] & 0x3f);
		}
	}

	if (0xff00 <= addr && addr <= 0xffef) {
//		data = ram[addr];
		data = 0xff;
	} else {
		data = rbank[addr >> BANK_SIZE][addr & ((1 << BANK_SIZE) - 1)];
	}

	// memory mapped i/o
	switch(addr & 0xffff) {
#ifdef USE_FD1
		case 0xff00:
		case 0xff01:
		case 0xff02:
		case 0xff03:
			// fdc
			if (IOPORT_USE_5FDD) {
				data = d_fdc5->read_io8(addr - 0xff00);
			}
			break;
		case 0xff04:
			// fdd drive select
			if (IOPORT_USE_5FDD) {
				data = d_fdd->read_io8(0);
			}
			break;

		case 0xff18:
		case 0xff19:
		case 0xff1a:
		case 0xff1b:
		case 0xff1c:
		case 0xff1d:
		case 0xff1e:
		case 0xff1f:
			// fdc
			if (IOPORT_USE_3FDD) {
				data = d_fdc3->read_io8(addr - 0xff18);

//				emu->out_debugf("fdc3r a:%04x d:%02x",addr,data);
			}
			break;
		case 0xff20:
			// fdd drive select
			if (IOPORT_USE_3FDD) {
				data = d_fdd->read_io8(0);

//				emu->out_debugf("fdd3r a:%04x d:%02x",addr,data);
			}
			break;
#endif
		case 0xff30:
		case 0xff31:
		case 0xff32:
		case 0xff33:
			// psg
			if (IOPORT_USE_PSG6) {
				data = d_psg->read_io8(addr - 0xff30);
//				emu->out_debugf("psgr a:%04x d:%02x",addr,data);
			}
			break;
#ifdef USE_RTC
		case 0xff3a:
			// rtc read
			if ((rtc_reg_sel & 0xc0) == 0x80) {
				int tim[8];
				emu->get_timer(tim);
				switch (rtc_reg_sel & 0x0f) {
				case 0:
					data = (tim[6] % 10);
					break;
				case 1:
					data = (tim[6] / 10);
					break;
				case 2:
					data = (tim[5] % 10);
					break;
				case 3:
					data = (tim[5] / 10);
					break;
				case 4:
					data = (tim[4] % 10);
					break;
				case 5:
					data = (tim[4] / 10);
					break;
				case 6:
					data = (tim[3] % 10);
					break;
				case 7:
					data = (tim[2] % 10);
					break;
				case 8:
					data = (tim[2] / 10);
					break;
				case 9:
					data = (tim[1] % 10);
					break;
				case 10:
					data = (tim[1] / 10);
					break;
				case 11:
					data = (tim[0] % 10);
					break;
				case 12:
					data = ((tim[0] / 10) % 10);
					break;
				}
			} else {
				data = rtc_buf & 0xff;
			}
			break;
#endif
		case 0xff3c:
		case 0xff3d:
		case 0xff3e:
		case 0xff3f:
			// extended LPT board
			if (IOPORT_USE_EXPIA) {
				data = d_pia_ex->read_io8(addr - 0xff3c);
			}
			break;

		case 0xff40:
		case 0xff41:
			// extended COM board
			if (IOPORT_USE_EXACIA) {
				data = d_acia_ex->read_io8(addr - 0xff40);
			}
			break;

		case 0xff70:
		case 0xff71:
		case 0xff72:
		case 0xff73:
		case 0xff74:
		case 0xff77:
		case 0xff78:
		case 0xff79:
		case 0xff7a:
		case 0xff7b:
		case 0xff7c:
		case 0xff7d:
		case 0xff7e:
		case 0xff7f:
			// psg9
			if (IOPORT_USE_PSG9) {
				data = d_psg9->read_io8(addr - 0xff70);
//				emu->out_debugf("psg9r a:%04x d:%02x",addr,data);
			}
			break;

		case 0xff75:
		case 0xff76:
			// psg9
			if (IOPORT_USE_PSG9) {
				data = d_psg9->read_io8(addr - 0xff70);
//				emu->out_debugf("psg9r a:%04x d:%02x",addr,data);
			}
			// kanji rom
			if (IOPORT_USE_KANJI) {
				data = d_kanji->read_io8(addr - 0xff75);
//				emu->out_debugf("kanjir a:%04x d:%02x",addr,data);
			}
			break;

		case 0xffc0:
		case 0xffc1:
		case 0xffc2:
		case 0xffc3:
			// pia
			data = d_pia->read_io8(addr - 0xffc0);
			break;

		case 0xffc4:
		case 0xffc5:
			// acia
			data = d_acia->read_io8(addr - 0xffc4);
			break;

		case 0xffc6:
		case 0xffc7:
			// crtc
			data = d_crtc->read_io8(addr - 0xffc6);
			break;

		case 0xffc8:
			// keyboard nmi
			data = d_key->read_io8(addr);
			break;
		case 0xffc9:
			// newon number
			data = config.dipswitch;
			break;
		case 0xffca:
			// timer irq
			data = d_timer->read_io8(addr);
			break;
		case 0xffcb:
			// light pen flag (cannot write)
			data = d_key->read_io8(addr);
			break;
		case 0xffd0:
			// mode select
			break;
		case 0xffd1:
			// trace counter
			break;
		case 0xffd2:
			// remote switch
			break;
		case 0xffd3:
			// music sel
			break;
		case 0xffd4:
			// time mask
			break;
		case 0xffd5:
			// light pen bl
			break;
		case 0xffd6:
			// interace sel
			break;
		case 0xffd7:
			// baud sel
			break;
		case 0xffd8:
			// color register sel
#ifdef _DEBUG_CRAM
			emu->out_debugf("cr c%02x",color_reg);
#endif
			data = color_reg & 0xbf;
			break;
		case 0xffe0:
			// keyboard scan data
			data = d_key -> read_io8(addr);
			break;
		case 0xffe8:
			// bank register
			break;
		case 0xffe9:
			// IG mode register
			break;
		case 0xffea:
			// IG en register
			break;
	}

#ifdef _DEBUG_RAM
	if (addr >= 0xff70 && addr <= 0xff7f) {
		uint8 ch = (data < 0x20 || data > 0x7f) ? 0x20 : data;
		emu->out_debugf("mr %04x=%02x %c",addr,data,ch);
	}
#endif

	return data;
}

#ifdef USE_CPU_REAL_MACHINE_CYCLE
void MEMORY::latch_address(uint32 addr, int *wait)
{

}
#endif

void MEMORY::write_io8(uint32 addr, uint32 data)
{

}

void MEMORY::write_signal(int id, uint32 data, uint32 mask)
{
	switch(id) {
		case SIG_MEMORY_PIA_PA:
			// from PIA portA
			mem_bank_reg1 = data & 0xce;
			change_memory_bank();
			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			if (now_reset == true) {
				// now on reset signal

				// reset igmode
				set_igmode(0);
				ig_enreg = 0;
			} else {
				// turn off reset signal

				// init memory map
				reset_membank();
				// stop trace counter
				reset_tracecounter();
			}
			break;
	}
}

void MEMORY::change_memory_bank()
{
//	mem_bank_reg1;	// from PIA port A
//	mem_bank_reg2;	// from 0xffe8

	// 0x0000 - 0x7fff
	if (mem_bank_reg2 & 0x08) {
		if (mem_bank_reg2 & 0x20) {
			// change to the extended ram both r/w
			SET_BANK(0x0000, 0x7fff, ram2 + 0x0000, ram2 + 0x0000);
		} else {
			// change to the extended ram only write
			SET_BANK(0x0000, 0x7fff, ram2 + 0x0000, ram + 0x0000);
		}
	} else {
		// change to standard
		SET_BANK(0x0000, 0x7fff, ram + 0x0000, ram + 0x0000);
	}

	// 0x0400 - 0x43ff (vram area)
	// exchange vram area to/from extended ram
	if ((mem_bank_reg2 & 0x0c) == 0x04 || (mem_bank_reg2 & 0x0c) == 0x08) {
		// select extend ram
		if (mem_bank_reg2 & 0x20) {
			// read/write
			SET_BANK(0x0400, 0x43ff, ram2 + 0x0400, ram2 + 0x0400);
		} else {
			// write only
			SET_BANK(0x0400, 0x43ff, ram2 + 0x0400, ram + 0x0400);
		}
	} else {
		// select standard ram
		SET_BANK(0x0400, 0x43ff, ram + 0x0400, ram + 0x0400);
	}

	// 0x8000 - 0xffff
	if ((mem_bank_reg2 & 0x50) == 0x50) {
		// change to the extended ram both r/w
		SET_BANK(0x8000, 0xffff, ram2 + 0x8000, ram2 + 0x8000);
	} else {
		//
		// decide the area to write
		if (mem_bank_reg2 & 0x10) {
			// change to the extended ram only write
			SET_BANK(0x8000, 0xffff, ram2 + 0x8000, NULL);
		} else {
			// change to the standard
			SET_BANK(0x8000, 0xffff, ram + 0x8000, NULL);
		}

		//
		// decide the area to read
		SET_BANK(0x8000, 0x9fff, NULL, ram + 0x8000);
		SET_BANK(0xa000, 0xffff, NULL, rom + 0x0000);
		if (IOPORT_USE_3FDD && rom1805_loaded) {
			// for 3inch compact floppy
			SET_BANK(0xf800, 0xffff, NULL, rom1805);
		}
		if (IOPORT_USE_5FDD && rom1802_loaded) {
			// for 5.25inch mini floppy
			SET_BANK(0xf800, 0xffff, NULL, rom1802);
		}

		if (!(mem_bank_reg1 & 0x02)) {
			if (mem_bank_reg1 & 0x40) {
				// r/w to ram (this means cannot read rom data !)
				SET_BANK(0xa000, 0xbfff, NULL, ram + 0xa000);
			}
		}
		if (!(mem_bank_reg1 & 0x04)) {
			if (mem_bank_reg1 & 0x80) {
				// r/w to ram (this means cannot read rom data !)
				SET_BANK(0xc000, 0xdfff, NULL, ram + 0xc000);
			}
		}
		if (!(mem_bank_reg1 & 0x08)) {
			if (mem_bank_reg1 & 0x80) {
				// r/w to ram (this means cannot read rom data !)
				SET_BANK(0xe000, 0xefff, NULL, ram + 0xe000);
			}
		}

		if (mem_bank_reg2 & 0x01) {
			// read from ram (this means cannot read rom data !)
			SET_BANK(0xf000, 0xfeff, NULL, ram + 0xf000);
		} else {
			// read from rom , write to ram
			SET_BANK(0xf000, 0xfeff, NULL, rom + 0x5000);
			if (IOPORT_USE_3FDD && rom1805_loaded) {
				// for 3inch compact floppy
				SET_BANK(0xf800, 0xfeff, NULL, rom1805);
			}
			if (IOPORT_USE_5FDD && rom1802_loaded) {
				// for 5.25inch mini floppy
				SET_BANK(0xf800, 0xfeff, NULL, rom1802);
			}
		}
		if (mem_bank_reg2 & 0x02) {
			// r/w to ram (this means cannot read rom data !)
			SET_BANK(0xfff0, 0xffff, NULL, ram + 0xfff0);
		} else {
			// read from rom , write to ram
			SET_BANK(0xfff0, 0xffff, NULL, rom + 0x5ff0);
		}
	}
}

// return vram area
uint8* MEMORY::get_vram() {
// vram is always the standard ram
//	return ((mem_bank_reg2 & 0x04) ? ram2 : ram );
	return ram;
}

void MEMORY::fetch_trace_counter() {
	pc = 0;
	pc_prev = TRACE_PC_FETCH;
//	emu->out_debugf(_T("trace_fetch: clk:%lld"), current_clock() % 900000);
}

void MEMORY::set_trace_counter() {
	// send NMI interrupt after few clocks (16 at least)
	register_event_by_clock(this, SIG_TRACE_COUNTER, 16, false, &trace_counter_id);
	pc = 0;
	pc_prev = TRACE_PC_WAITING;
//	emu->out_debugf(_T("trace_start: clk:%lld"), current_clock() % 900000);
}

void MEMORY::clear_tracecounter()
{
	// clear NMI
	d_board->write_signal(SIG_CPU_NMI, 0, SIG_NMI_TRACE_MASK);
//	emu->out_debugf(_T("trace NMI off: clk:%lld"), current_clock() % 900000);
	pc_prev = TRACE_PC_IDLE;
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void MEMORY::event_callback(int event_id, int err)
{
	if (event_id == SIG_TRACE_COUNTER) {
		// event stop
//		cancel_event(trace_counter_id);
		trace_counter_id = -1;

		pc = d_cpu->get_pc();

#if 0
		if (pc == pc_prev || pc_prev == TRACE_PC_WAITING) {
			// inhibit the NMI interrupt until next opcode is executed.
			pc_prev = pc;
			// next event
			register_event(this, SIG_TRACE_COUNTER, 2, false, &trace_counter_id);
//			emu->out_debugf(_T("trace_fire: clk:%lld pc:%04x next"), current_clock() % 900000, pc);
		} else
#endif
		{
			// send NMI interrupt
			d_board->write_signal(SIG_CPU_NMI, SIG_NMI_TRACE_MASK, SIG_NMI_TRACE_MASK);
//			emu->out_debugf(_T("trace_fire: clk:%lld pc:%04x NMI on"), current_clock() % 900000, pc);
			pc_prev = TRACE_PC_FIRED;
		}
	}
}

// ----------------------------------------------------------------------------

void MEMORY::update_config()
{
}

// ----------------------------------------------------------------------------

void MEMORY::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 1;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	memcpy(vm_state.ram,  ram,  sizeof(ram));
	memcpy(vm_state.ram2, ram2, sizeof(ram2));
	memcpy(vm_state.color_ram, color_ram, sizeof(color_ram));
	memcpy(vm_state.ig_ram, ig_ram, sizeof(ig_ram));

	vm_state.mem_bank_reg1 = mem_bank_reg1;
	vm_state.mem_bank_reg2 = mem_bank_reg2;

	vm_state.trace_counter_id = trace_counter_id;
	vm_state.pc = pc;
	vm_state.pc_prev = pc_prev;

	// save config
	vm_state.dipswitch = config.dipswitch;
	vm_state.fdd_type = config.fdd_type;
	vm_state.io_port = config.io_port;

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void MEMORY::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	// copy values
	memcpy(ram,  vm_state.ram,  sizeof(ram));
	memcpy(ram2, vm_state.ram2, sizeof(ram2));
	memcpy(color_ram, vm_state.color_ram, sizeof(color_ram));
	memcpy(ig_ram, vm_state.ig_ram, sizeof(ig_ram));

	mem_bank_reg1 = vm_state.mem_bank_reg1;
	mem_bank_reg2 = vm_state.mem_bank_reg2;

	trace_counter_id = vm_state.trace_counter_id;
	pc = vm_state.pc;
	pc_prev = vm_state.pc_prev;

	// calc values
//	write_data8(0xffd0, ram[0xffd0]);	// mode sel
	// rs bit on mode_sel is set in COMM device. 
	d_disp->write_io8(0xffd0, ram[0xffd0]);
	d_crtc->write_signal(SIG_HD46505_CHR_CLOCKS, CPU_CLOCKS * ((ram[0xffd0] & 0x80) ? 2 : 1), 0xffffffff);

	write_data8(0xffd6, ram[0xffd6]);	// interace sel
	write_data8(0xffd8, ram[0xffd8]);	// color reg

	write_data8(0xffea, ram[0xffea]);	// IG en register

	// load config
	config.dipswitch = vm_state.dipswitch;
	config.fdd_type = vm_state.fdd_type;
	config.io_port = vm_state.io_port;

	change_memory_bank();
}
