/** @file ym2203.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.09.15-

	@note
	Modified by Sasaji at 2012.06.20

	@brief [ AY-3-8912 / YM2203 / YM2608 ]
*/

#include "ym2203.h"

void YM2203::initialize()
{
#ifdef HAS_YM2608
	chip = new FM::OPNA;
#else
	chip = new FM::OPN;
#endif
	register_vline_event(this);
	mute = false;
}

void YM2203::release()
{
	delete chip;
}

void YM2203::reset()
{
	chip->Reset();
	chip->SetReg(0x27, 0);
#if defined(_X1TWIN) || defined(_X1TURBO)
	chip->SetReg(0x2e, 0);
#endif
	port[0].first = port[1].first = true;
}

#ifdef HAS_YM2608
#define amask 3
#else
#define amask 1
#endif

void YM2203::write_io8(uint32 addr, uint32 data)
{
	switch(addr & amask) {
	case 0:
#ifdef HAS_AY_3_8912
		ch = data & 0x0f;
#else
		ch = data;
		// write dummy data for prescaler
		if(0x2d <= ch && ch <= 0x2f) {
			chip->SetReg(ch, 0);
		}
#endif
		break;
	case 1:
		if(ch == 7) {
			mode = data;
		}
		else if(ch == 14 || ch == 15) {
			int p = ch - 14;
			if(port[p].wreg != data || port[p].first) {
				write_signals(&port[p].outputs, data);
				port[p].wreg = data;
				port[p].first = false;
			}
		}
		// don't write again for prescaler
		if(!(0x2d <= ch && ch <= 0x2f)) {
			chip->SetReg(ch, data);
#ifndef HAS_AY_3_8912
			update_interrupt();
#endif
		}
		break;
#ifdef HAS_YM2608
	case 2:
		ch1 = data1 = data;
		break;
	case 3:
		chip->SetReg(0x100 | ch1, data);
		data1 = data;
		update_interrupt();
		break;
#endif
	}
}

uint32 YM2203::read_io8(uint32 addr)
{
	switch(addr & amask) {
#ifndef HAS_AY_3_8912
	case 0:
		return chip->ReadStatus();
#endif
	case 1:
		if(ch == 14) {
#if defined(_PC98DO) || defined(_PC8801MA)
			return port[0].rreg;
#else
			return (mode & 0x40) ? port[0].wreg : port[0].rreg;
#endif
		}
		else if(ch == 15) {
#if defined(_PC98DO) || defined(_PC8801MA)
			return port[1].rreg;
#else
			return (mode & 0x80) ? port[1].wreg : port[1].rreg;
#endif
		}
		return chip->GetReg(ch);
#ifdef HAS_YM2608
	case 2:
		return chip->ReadStatusEx();
	case 3:
		if(ch1 == 8) {
			return chip->GetReg(0x100 | ch1);
		}
		return data1;
#endif
	}
	return 0xff;
}

void YM2203::write_signal(int id, uint32 data, uint32 mask)
{
	if(id == SIG_YM2203_PORT_A) {
		port[0].rreg = (port[0].rreg & ~mask) | (data & mask);
	}
	else if(id == SIG_YM2203_PORT_B) {
		port[1].rreg = (port[1].rreg & ~mask) | (data & mask);
	}
	else if(id == SIG_YM2203_MUTE) {
		mute = ((data & mask) != 0);
	}
	else if (id == SIG_CPU_RESET) {
		now_reset = (data & mask) ? true : false;
		reset();
	}
}

void YM2203::event_vline(int v, int clock)
{
	chip->Count(usec_per_vline);
#ifndef HAS_AY_3_8912
	update_interrupt();
#endif
}

#ifndef HAS_AY_3_8912
void YM2203::update_interrupt()
{
	if(chip->ReadIRQ()) {
		write_signals(&outputs_irq, 0);
		write_signals(&outputs_irq, 0xffffffff);
	}
}
#endif

void YM2203::mix(int32* buffer, int cnt)
{
	if (sound_volume != 0) {
#ifdef USE_EMU_INHERENT_SPEC
		memset(buffer_tmp, 0, sizeof(buffer_tmp));
		if (cnt > 20) cnt = 20;
		if(cnt > 0 && !mute) {
			chip->Mix(buffer_tmp, cnt);
		}
		for(int i=0; i<(cnt * 2); i++) {
			*buffer++ += (buffer_tmp[i] * sound_volume / 100);
		}
#else
		if(cnt > 0 && !mute) {
			chip->Mix(buffer, cnt);
		}
#endif
	}
}

void YM2203::init(int rate, int clock, int samples, int volf, int volp, int ptn)
{
#ifdef HAS_YM2608
	_TCHAR app_path[_MAX_PATH];
	emu->application_path(app_path);
	chip->Init(clock, rate, false, app_path);
#else
	chip->Init(clock, rate, false, NULL);
#endif
	chip->SetVolumeFM(volf);
	chip->SetVolumePSG(volp, ptn);
#ifdef USE_EMU_INHERENT_SPEC
	sound_volume = 0;
#endif
}

void YM2203::update_timing(int new_clocks, double new_frames_per_sec, int new_lines_per_frame)
{
	usec_per_vline = (int)(1000000.0 / new_frames_per_sec / (double)new_lines_per_frame + 0.5);
}

#ifdef USE_EMU_INHERENT_SPEC
void YM2203::set_volume(int volume)
{
	sound_volume = volume;
}
#endif

// ----------------------------------------------------------------------------
#ifdef USE_EMU_INHERENT_SPEC
void YM2203::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 2;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	// only save PSG register
	vm_state.ch = ch;
	vm_state.mode = mode;
	for(int i=0; i<14; i++) {
		vm_state.v2.reg[i] = chip->GetReg(i);
	}
	for(int i=0; i<2; i++) {
		vm_state.v2.port[i].wreg = port[i].wreg;
		vm_state.v2.port[i].rreg = port[i].rreg;
		vm_state.v2.port[i].first = port[i].first ? 1 : 0;
	}

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void YM2203::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	// only load PSG register
	ch = vm_state.ch;
	mode = vm_state.mode;
	if (vm_state_i.version >= 2) {
		for(int i=0; i<14; i++) {
			 chip->SetReg(i, vm_state.v2.reg[i]);
		}
		for(int i=0; i<2; i++) {
			port[i].wreg = vm_state.v2.port[i].wreg;
			port[i].rreg = vm_state.v2.port[i].rreg;
			port[i].first = vm_state.v2.port[i].first ? true : false;
		}
	} else {
		for(int i=0; i<13; i++) {
			 chip->SetReg(i, vm_state.v1.reg[i]);
		}
		for(int i=0; i<2; i++) {
			port[i].wreg = vm_state.v1.port[i].wreg;
			port[i].rreg = vm_state.v1.port[i].rreg;
			port[i].first = vm_state.v1.port[i].first ? true : false;
		}
	}
}
#endif